﻿--[[
	Copyright (c) 2012, Lars "Goldpaw" Norberg
	
	Web: http://www.friendlydruid.com
	Contact: goldpaw@friendlydruid.com
	
	All rights reserved
]]--
local _; local addon,ns = ...

if (select(4, GetBuildInfo())) < 50001 then
	print(("gBags v3 requires WoW patch 5.0.1 or higher, and you only have %s, bailing out!"):format((GetBuildInfo())))
	return
end

if (IsAddOnLoaded("gUI-v3")) then
	print(GetAddOnMetadata(addon, "Title") .. ": Detected gUI™, bailing out!")
	return
end

local gUI = LibStub("gCore-4.0"):NewAddon(addon, "gDB-2.0", "gMedia-4.0", "gOptionsMenu-1.0", "gFrameHandler-2.0")
if not(gUI) then return end

gUI:SetDefaultModuleState(false) -- allow modules to be manually enabled, preventing them from firing OnInit prematurely
gUI:SetDefaultModuleLibraries("gOptionsMenu-1.0", "gFrameHandler-2.0") -- give all our modules default access to these libraries

local rawget = rawget
local setmetatable = setmetatable
local floor, max = math.floor, math.max
local type, tonumber = type, tonumber
local pairs, select, unpack = pairs, select, unpack
local format, gsub = string.format, string.gsub
local tconcat, tsort, tinsert, tremove = table.concat, table.sort, table.insert, table.remove
local strsplit = string.split

local GetContainerNumFreeSlots = GetContainerNumFreeSlots
local GetContainerNumSlots = GetContainerNumSlots
local GetGameTime = GetGameTime
local GetMoney = GetMoney
local InCombatLockdown = InCombatLockdown
local IsAddOnLoaded = IsAddOnLoaded
local ReloadUI = ReloadUI
local StaticPopup_Show = StaticPopup_Show
local StaticPopupDialogs = StaticPopupDialogs
local BNGetFriendInfo = BNGetFriendInfo
local BNGetNumFriends = BNGetNumFriends
local BNGetToonInfo = BNGetToonInfo
local GetAddOnMetadata = GetAddOnMetadata
local GetBuildInfo = GetBuildInfo
local GetGuildRosterInfo = GetGuildRosterInfo
local GetNumFriends = GetNumFriends
local GetNumGuildMembers = GetNumGuildMembers
local GetNumGroupMembers =  GetNumGroupMembers
local GetNumSpecGroups = GetNumSpecGroups
local GetNumSpecializations = GetNumSpecializations
local GetNumWorldPVPAreas = GetNumWorldPVPAreas
local GetRealZoneText = GetRealZoneText
local GetSpecializationInfo = GetSpecializationInfo
local GetSpecializationRole = GetSpecializationRole
local GetQuestDifficultyColor = GetQuestDifficultyColor
local GetWorldPVPAreaInfo = GetWorldPVPAreaInfo
local IsInInstance = IsInInstance

local gABT = LibStub("gActionButtons-3.0")
local gOM = LibStub("gOptionsMenu-1.0")
local L = LibStub("gLocale-2.0"):GetLocale(addon)
local C = gUI:NewDataBase("colors")
local F = gUI:NewDataBase("functions")
local M = function(folder, file) return gUI:GetMedia(folder, file) end -- our magic proxy function to retrieve gUI's media
local noop = noop or gUI.noop
local defaults, db
local addonQueue = {} -- for the skinner functionality, and other stuff

local TEST_MODE = false 

local RegisterMenuWidgets


--------------------------------------------------------------------------------------------------
--		default settings
--------------------------------------------------------------------------------------------------
do
	defaults = {
		-- version control
		initialized = false; -- will be true when user has run through install tutorial once
		version = 0; -- version of gUIv3 (this is NOT 3, but rather whatever is listed in the .toc file)
		build = 0; -- build number as reported by the Curse Packager
		temporaryDisabled = false;
		
		-- user configurable colors and values. not yet implemented.
		glossAlpha = 1/3; 
		shadeAlpha = 1/2;
		panelAlpha = 75/100; -- alpha for background panels, tooltips, etc
		overlayAlpha = 1/5; -- alpha for overlay panels 
		gold = 0;

		-- enabled modules (0 = never, 1 = load unless autodisabled, 2 = load always)
		-- not all of these can be directly modifed by the user, even though they have a saved setting
		enabledModules = { ["Bags"] = 1 };
	}
end

--------------------------------------------------------------------------------------------------
--		variables available to sub-modules and external addons
--------------------------------------------------------------------------------------------------
do
	gUI:SetAttribute("dummy", CreateFrame("Frame")); gUI:GetAttribute("dummy"):Hide() -- taintfree way of hiding objects
	gUI:SetAttribute("parent", CreateFrame("Frame", nil, UIParent, "SecureHandlerStateTemplate")) -- top level frame hidden in pet battles
	RegisterStateDriver(gUI:GetAttribute("parent"), "visibility", "[petbattle] hide; show")
	
	gUI:SetAttribute("vehiclehider", CreateFrame("Frame", nil, gUI:GetAttribute("parent"), "SecureHandlerStateTemplate")) -- 2nd level frame hidden in vehicles
	RegisterStateDriver(gUI:GetAttribute("vehiclehider"), "visibility", "[vehicleui] hide; show")
	
	gUI:SetAttribute("savedvariablesglobal", "gBags_DB") -- saved variables global name, as set in the .toc file
	gUI:SetAttribute("title", GetAddOnMetadata(addon, "Title")) -- display name of the addon
	gUI:SetAttribute("version", GetAddOnMetadata(addon, "Version")) -- addon version
	gUI:SetAttribute("curseversion", GetAddOnMetadata(addon, "X-Curse-Packaged-Version")) -- curse packaged version/build
	gUI:SetAttribute("realm", GetRealmName()) -- current realm name
	gUI:SetAttribute("name", UnitName("player")) -- player name
	gUI:SetAttribute("class", select(2, UnitClass("player"))) -- player class file name
	gUI:SetAttribute("classID", select(3, UnitClass("player"))) -- player class ID
	gUI:SetAttribute("class-local", UnitClass("player")) -- localized player class display name
	gUI:SetAttribute("faction", UnitFactionGroup("unit")) -- english player faction name
	gUI:SetAttribute("faction-local", select(2, UnitFactionGroup("unit"))) -- localized player faction name
end

--------------------------------------------------------------------------------------------------
--		media library
--------------------------------------------------------------------------------------------------
do
	gUI:SetDefaultMediaPath(([[Interface\AddOns\%s\media\]]):format(addon)) -- retrieve the actual filepath no matter the addon name
	-------------------------------------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------------------------------------
	--
	-- 	All the gUI™ media, only registered in the main addon object, but available to all submodules
	-- 		*the 'gUI™'-prefixes indicates that I made that media and it may not be used without written permission
	-- 		*taking bets on when I'll go completely mental from all the 'gUI™'-names
	--
	-------------------------------------------------------------------------------------------------------------------------------
	-------------------------------------------------------------------------------------------------------------------------------
	
	-- backgrounds
	gUI:NewMedia("Background", "gUI™ Gloss", [[backgrounds\Gloss64x64.tga]])
	gUI:NewMedia("Background", "gUI™ LargeGloss", [[backgrounds\Gloss256x256.tga]])
	gUI:NewMedia("Background", "gUI™ Shade", [[backgrounds\Shade64x64.tga]])
	gUI:NewMedia("Background", "gUI™ LargeShade", [[backgrounds\Shade256x256.tga]])
	gUI:NewMedia("Background", "gUI™ MildSatin", [[backgrounds\MildSatin256x64.tga]])
	gUI:NewMedia("Background", "gUI™ UnitShader", [[backgrounds\UnitShader256x64.tga]])
	gUI:NewMedia("Background", "gUI™ VoidStorage", [[backgrounds\VoidStorage32x32.tga]])
	
	-- borders
	gUI:NewMedia("Border", "gUI™ GlowBorder", [[borders\gUI-GlowBorder128x16.tga]])
	gUI:NewMedia("Border", "gUI™ GlowBorderLarge", [[borders\gUI-GlowBorder256x32.tga]])
	gUI:NewMedia("Border", "gUI™ PixelBorder", [[borders\gUI-PixelBorder128x16.tga]])
	gUI:NewMedia("Border", "gUI™ PixelBorderIndented", [[borders\gUI-PixelBorder128x16-2px-Indented.tga]])

	-- buttons
	gUI:NewMedia("Button", "gUI™ BagIcon", [[buttons\BagIcon32x32.tga]])
	gUI:NewMedia("Button", "gUI™ BagIconDisabled", [[buttons\BagIconDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ BagIconHighlight", [[buttons\BagIconHighlight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ BagRestackIcon", [[buttons\BagRestackIcon64x32.tga]])
	gUI:NewMedia("Button", "gUI™ BagRestackIconDisabled", [[buttons\BagRestackIconDisabled64x32.tga]])
	gUI:NewMedia("Button", "gUI™ BagRestackIconHighlight", [[buttons\BagRestackIconHighlight64x32.tga]])
	
	-- buttons: arrows
	gUI:NewMedia("Button", "gUI™ ArrowBottom", [[buttons\ArrowBottom32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowBottomDisabled", [[buttons\ArrowBottomDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowBottomHighlight", [[buttons\ArrowBottomHighlight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowTop", [[buttons\ArrowTop32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowTopDisabled", [[buttons\ArrowTopDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowTopHighlight", [[buttons\ArrowTopHighlight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowFirst", [[buttons\ArrowFirst32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowFirstDisabled", [[buttons\ArrowFirstDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowFirstHighlight", [[buttons\ArrowFirstHighlight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowLast", [[buttons\ArrowLast32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowLastDisabled", [[buttons\ArrowLastDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowLastHighlight", [[buttons\ArrowLastHighlight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowRight", [[buttons\ArrowRight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowRightDisabled", [[buttons\ArrowRightDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowRightHighlight", [[buttons\ArrowRightHighlight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowLeft", [[buttons\ArrowLeft32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowLeftDisabled", [[buttons\ArrowLeftDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowLeftHighlight", [[buttons\ArrowLeftHighlight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowUp", [[buttons\ArrowUp32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowUpDisabled", [[buttons\ArrowUpDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowUpHighlight", [[buttons\ArrowUpHighlight32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowDown", [[buttons\ArrowDown32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowDownDisabled", [[buttons\ArrowDownDisabled32x32.tga]])
	gUI:NewMedia("Button", "gUI™ ArrowDownHighlight", [[buttons\ArrowDownHighlight32x32.tga]])

	-- buttons: window navigation
	gUI:NewMedia("Button", "gUI™ CloseButton", [[buttons\gUI-CloseButton32x32-Up.tga]])
	gUI:NewMedia("Button", "gUI™ CloseButtonDown", [[buttons\gUI-CloseButton32x32-Down.tga]])
	gUI:NewMedia("Button", "gUI™ CloseButtonDisabled", [[buttons\gUI-CloseButton32x32-Disabled.tga]])
	gUI:NewMedia("Button", "gUI™ CloseButtonHighlight", [[buttons\gUI-CloseButton32x32-Highlight.tga]])
	gUI:NewMedia("Button", "RefreshArrow", [[buttons\RefreshArrow32x32.tga]])
	gUI:NewMedia("Button", "StopSign", [[buttons\StopSign32x32.tga]])

	-- icons
	gUI:NewMedia("Icon", "Cog", [[icons\Cog32x32.tga]])

	-- statusbars
	gUI:NewMedia("Statusbar", "gUI™ ProgressBar", [[statusbars\progressbar.tga]])
	
	-- v3 fonts
	gUI:NewMedia("Font", "Oswald Light", [[fonts\Oswald-Light.ttf]])
	gUI:NewMedia("Font", "Oswald Regular", [[fonts\Oswald-Regular.ttf]])
	gUI:NewMedia("Font", "Oswald Bold", [[fonts\Oswald-Bold.ttf]])
	gUI:NewMedia("Font", "PT Sans Narrow", [[fonts\PT Sans Narrow.ttf]])
	gUI:NewMedia("Font", "PT Sans Narrow Bold", [[fonts\PT Sans Narrow Bold.ttf]])

	-- backdrops
	gUI:NewMedia("Backdrop", "PixelBorder", {
		edgeFile = M("Border", "gUI™ PixelBorder"); 
		edgeSize = 8;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})

	gUI:NewMedia("Backdrop", "PixelBorder-Blank", {
		bgFile = M("Background", "Blank"); 
		edgeFile = M("Border", "gUI™ PixelBorder"); 
		edgeSize = 8;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})

	gUI:NewMedia("Backdrop", "SimpleBorder", {
		bgFile = M("Background", "Blank"); 
		edgeFile = M("Background", "Blank"); 
		edgeSize = 1;
		insets = { 
			bottom = -1; 
			left = -1; 
			right = -1; 
			top = -1; 
		};
	})

	gUI:NewMedia("Backdrop", "PixelBorder-Blank-Indented", {
		bgFile = M("Background", "Blank"); 
		edgeFile = M("Border", "gUI™ PixelBorderIndented"); 
		edgeSize = 8;
		insets = { 
			bottom = 2; 
			left = 2; 
			right = 2; 
			top = 2; 
		};
	})

	gUI:NewMedia("Backdrop", "PixelBorder-Satin", {
		bgFile = M("Background", "gUI™ MildSatin"); 
		edgeFile = M("Border", "gUI™ PixelBorder"); 
		edgeSize = 8;
		insets = { 
			bottom = 2; 
			left = 2; 
			right = 2; 
			top = 2; 
		};
	})

	gUI:NewMedia("Backdrop", "PixelBorder-Satin-Indented", {
		bgFile = M("Background", "gUI™ MildSatin"); 
		edgeFile = M("Border", "gUI™ PixelBorderIndented"); 
		edgeSize = 8;
		insets = { 
			bottom = 2; 
			left = 2; 
			right = 2; 
			top = 2; 
		};
	})

	gUI:NewMedia("Backdrop", "TargetBorder", {
		bgFile = nil; 
		edgeFile = M("Background", "Blank"); 
		edgeSize = 2;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})
	
	gUI:NewMedia("Backdrop", "StatusBarBorder", {
		bgFile = M("Statusbar", "gUI™ StatusBar"); 
		edgeFile = M("Border", "gUI™ PixelBorder"); 
		edgeSize = 8;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})
	
	
	gUI:NewMedia("Backdrop", "TinyGlow", {
		edgeFile = M("Border", "gUI™ GlowBorder"); 
		edgeSize = 3;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})

	gUI:NewMedia("Backdrop", "SmallGlow", {
		edgeFile = M("Border", "gUI™ GlowBorder"); 
		edgeSize = 5;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})

	gUI:NewMedia("Backdrop", "Glow", {
		edgeFile = M("Border", "gUI™ GlowBorder"); 
		edgeSize = 16;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})

	gUI:NewMedia("Backdrop", "HugeGlow", {
		edgeFile = M("Border", "gUI™ GlowBorderLarge"); 
		edgeSize = 32;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})

	gUI:NewMedia("Backdrop", "ItemButton", {
		bgFile = M("Background", "gUI™ ItemButton");
		edgeFile = M("Border", "gUI™ PixelBorder"); 
		edgeSize = 8;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})

	gUI:NewMedia("Backdrop", "VoidStorageButton", {
		bgFile = M("Background", "gUI™ VoidStorage");
		edgeFile = M("Border", "gUI™ PixelBorder"); 
		edgeSize = 8;
		insets = { 
			bottom = 0; 
			left = 0; 
			right = 0; 
			top = 0; 
		};
	})
end

--------------------------------------------------------------------------------------------------
--		color library
--------------------------------------------------------------------------------------------------
-- ToDo:
-- 	- remove deprecated entries
-- 	- remove all r = 1, g = 1 etc
-- 	- add functions to easier set and retrieve main UI colors
do
	-- various
	C["realm"] = { 1.00; 1.00; 1.00; }
	C["guild"] = { 1.00; 1.00; 0.70; }
	C["error"] = { 1.00; 0.00; 0.00; }

	-- chat
	C["GeneralChat"] = { 0.6, 0.6, 1.0 }
	C["TradeChat"] = { 0.4, 0.4, 0.8 }
	C["RaidChat"] = { 1.0, 0.28, 0.04 } -- same as the original RaidLeader
	C["RaidLeader"] = { 1.0, 0.82, 0.0 }

	-- bag module
	C["bank"] = { 0.35, 0.35, 0.75, 1.00 }

	-- nameplate and tooltip stuff
	C["shield"] = { 0.95; 0.35; 0.05; 0.75 }
	C["rare"] = { 0.00; 0.44; 0.87; 0.75 }; -- rare mobs, but same as rare quality loot
	C["boss"] = { r = 1.00; g = 0.00; b = 0.00 }
	-- C["boss"] = { r = 1.00; g = 0.00; b = 0.00 }

	-- threat coloring for unitframe (or other) borders
	C["target"] = { 0.95; 0.95; 0.65; 0.85 }
	C["threat"] = { 0.95; 0.05; 0.05; 0.75 }
	C["lessthreat"] = { 0.95; 0.65; 0.05; 0.75 }

	-- will always return the 'correct' color, 
	--	as long as the percentage you insert is from 0 to 100. 
	-- (less than 0 will return 0, and more than 100 returns 100)
	-- e.g C["durability"][28.5] will return the 25% color; C["durability"][25]
	C["durability"] = {
		[0] = { 1.00; 0.00; 0.00; };
		[1] = { 1.00; 0.30; 0.00; };
		[10] = { 1.00; 0.60; 0.00; };
		[25] = { 1.00; 1.00; 0.00; };
		[50] = { 1.00; 1.00; 0.60; };
		[100] = { 1.00; 1.00; 1.00; };
	}
	setmetatable(C["durability"], { __index = function(t, i) 
		if not(tonumber(i)) then return rawget(t, 0) end
		
		if (i < 1) then return rawget(t, 0)
		elseif (i < 10) then return rawget(t, 1)
		elseif (i < 25) then return rawget(t, 10)
		elseif (i < 50) then return rawget(t, 25)
		elseif (i < 100) then return rawget(t, 50)
		else return rawget(t, 100)
		end
	end })

	-- duration coloring for timers
	C["duration"] = {
	}

	-- panels and skinned frames & buttons
	C["index"] = { 1.00; 1.00; 1.00 }
	C["value"] = { 1.00; 0.82; 0.00 }
	C["dead"] = { 0.50; 0.50; 0.50 }
	C["disabled"] = { 0.40; 0.40; 0.40 }
	C["background"] = { 0.00; 0.00; 0.00 } -- 0
	C["backdrop"] = { 0.55; 0.55; 0.55 } -- 0.55
	C["border"] = { 0.15; 0.15; 0.15 } -- 0.15
	C["overlay"] = { 0.25; 0.25; 0.25 } -- 0.25
	C["hovercolor"] = { 0.35; 0.35; 0.35 } 
	C["hoverbordercolor"] = { 0.55; 0.55; 0.55 }
	C["shadow"] = { 0.00; 0.00; 0.00; 1 } 

	-- LFD role coloring. Experimental.
	C["role"] = {
		["tank"] = { 0.00; 0.25; 0.45 };
		["dps"] = { 0.45; 0.00; 0.00 };
		["heal"] = { 0.00; 0.45; 0.00 };
	}

	-- actionbutton borders
	-- C["normal"] = { r = 0.15; g = 0.15; b = 0.15; } -- normal state, when nothing else applies

	-- actionbutton icons
	C["equipped"] = { r = 0.10; g = 0.65; b = 0.10; } -- equipped items
	C["range"] = { r = 0.85; g = 0.10; b = 0.10; } -- target is out of range
	C["mana"] = { r = 0.10; g = 0.10; b = 0.85; } -- you haven't enough mana
	C["usable"] = { r = 1.00; g = 1.00; b = 1.00; } -- an item or ability is usable 
	C["unusable"] = { r = 0.35; g = 0.35; b = 0.35; } -- an item or ability is not usable

	-- actionbutton overlay colors
	-- C["gloss"] = { r = 1.00; g = 1.00; b = 1.00; a = 1/3 } -- gloss layer
	-- C["shade"] = { r = 0.00; g = 0.00; b = 0.00; a = 1/3 } -- shade layer

	-- the following only apply if panels are used instead of textures for actionbuttons
	-- which always is the case in gUI™ and Laevia™
	C["checked"] = { r = 1.00; g = 0.90; b = 0.00; } -- when checked (active stance, etc)
	C["flash"] = { r = 1.00; g = 0.00; b = 0.00; } -- hell if I know
	C["hover"] = { r = 1.00; g = 1.00; b = 1.00; } -- when hovering over a button
	C["pushed"] = { r = 1.00; g = 0.82; b = 0.00; } -- when pushing a button

	-- damage feedback for unitframes
	C["feedbackcolors"] = {
		DAMAGE = { 0.69; 0.31; 0.31 };
		CRUSHING = { 0.69; 0.31; 0.31 };
		CRITICAL = { 0.69; 0.31; 0.31 };
		GLANCING = { 0.69; 0.31; 0.31 };
		STANDARD = { 0.84; 0.75; 0.65 };
		IMMUNE = { 0.84; 0.75; 0.65 };
		ABSORB = { 0.84; 0.75; 0.65 };
		BLOCK = { 0.84; 0.75; 0.65 };
		RESIST = { 0.84; 0.75; 0.65 };
		MISS = { 0.84; 0.75; 0.65 };
		HEAL = { 0.33; 0.59; 0.33 };
		CRITHEAL = { 0.33; 0.59; 0.33 };
		ENERGIZE = { 0.31; 0.45; 0.63 };
		CRITENERGIZE = { 0.31; 0.45; 0.63 };
	}

	-- our own class colors
	C.RAID_CLASS_COLORS = setmetatable({
		["HUNTER"] = { r = 0.67, g = 0.83, b = 0.45, colorStr = "ffabd473" },
		["WARLOCK"] = { r = 0.58, g = 0.51, b = 0.79, colorStr = "ff9482c9" },
		-- ["PRIEST"] = { r = 1.0, g = 1.0, b = 1.0, colorStr = "ffffffff" },
		["PALADIN"] = { r = 0.96, g = 0.55, b = 0.73, colorStr = "fff58cba" },
		["MAGE"] = { r = 0.41, g = 0.8, b = 0.94, colorStr = "ff69ccf0" },
		["ROGUE"] = { r = 1.0, g = 0.96, b = 0.41, colorStr = "fffff569" },
		["DRUID"] = { r = 1.0, g = 0.49, b = 0.04, colorStr = "ffff7d0a" },
		["SHAMAN"] = { r = 0.0, g = 0.44, b = 0.87, colorStr = "ff0070de" },
		["WARRIOR"] = { r = 0.78, g = 0.61, b = 0.43, colorStr = "ffc79c6e" },
		["DEATHKNIGHT"] = { r = 0.77, g = 0.12 , b = 0.23, colorStr = "ffc41f3b" },
		["MONK"] = { r = 0.0, g = 1.00 , b = 0.59, colorStr = "ff00ff96" },
	
		-- ["MONK"] = { r = 0.00; g = 0.78; b = 0.48, colorStr = "ff00c77b" }; -- MMO-Champion #00C77B (April 12th 2012)
		["PRIEST"] = { r = 0.90; g = 0.90; b = 0.90, colorStr = "ffe5e5e5" }, -- keep this slightly gray
		["UNKNOWN"] = { r = 0.60; g = 0.60; b = 0.60, colorStr = "ff999999" }
	}, { __index = RAID_CLASS_COLORS })

	-- reputation coloring
	C.FACTION_BAR_COLORS = setmetatable({
		[1] = { r = 0.75; g = 0.10; b = 0.00 };
		[2] = { r = 0.75; g = 0.10; b = 0.00 };
		[3] = { r = 0.85; g = 0.40; b = 0.00 };
		[4] = { r = 0.90; g = 0.70; b = 0.10 };
		[5] = { r = 0.00; g = 0.60; b = 0.10 };
		[6] = { r = 0.00; g = 0.60; b = 0.10 };
		[7] = { r = 0.00; g = 0.60; b = 0.10 };
		[8] = { r = 0.00; g = 0.60; b = 0.10 };
	}, { __index = FACTION_BAR_COLORS })

	-- threat colors for statusbars and text
	C["THREAT_STATUS_COLORS"] = {
		[0] = { 1/10, 1/10, 1/10 }; -- this is where anybody not being the tank should be
		[1] = { 1, 1, 0.47 }; -- this is a dps dimwit with no comprehension of anything but dps meters
		[2] = { 1, 0.6, 0 }; -- this is when the tank is losing aggro due to overnuking dps dimwits
		[3] = { 1 * 1/3, 0, 0 }; -- this is where only the tank should be
	}

	-- health and power bars
	C["health"] = { C.FACTION_BAR_COLORS[5].r * 2/3, C.FACTION_BAR_COLORS[5].g * 2/3, C.FACTION_BAR_COLORS[5].b * 2/3, 1.00 }

	-- combatlog GUIS filter coloring
	C["combatlog"] = {
		["defaults"] = {
			["spell"] = { a = 1.00, r = C.index[1], g = C.index[2], b = C.index[3] };
			["damage"] = { a = 1.00, r = C.value[1], g = C.value[2], b = C.value[3] };
		};
		["schoolColoring"] = {
			[SCHOOL_MASK_NONE] = { a = 1.00, r = 1.00, g = 1.00, b = 1.00 };
			[SCHOOL_MASK_PHYSICAL] = { a = 1.00, r = 1.00, g = 1.00, b = 0.00 };
			[SCHOOL_MASK_HOLY] = { a = 1.00, r = 1.00, g = 0.90, b = 0.50 };
			[SCHOOL_MASK_FIRE] = { a = 1.00, r = 1.00, g = 0.50, b = 0.00 };
			[SCHOOL_MASK_NATURE] = { a = 1.00, r = 0.30, g = 1.00, b = 0.30 };
			[SCHOOL_MASK_FROST] = { a = 1.00, r = 0.50, g = 1.00, b = 1.00 };
			[SCHOOL_MASK_SHADOW] = { a = 1.00, r = 0.50, g = 0.50, b = 1.00 };
			[SCHOOL_MASK_ARCANE] = { a = 1.00, r = 1.00, g = 0.50, b = 1.00 };
		};
		["unitColoring"] = {
			[COMBATLOG_FILTER_MINE] = { a = 1, r = 0.40, g = 0.50, b = 1.00 };
			[COMBATLOG_FILTER_MY_PET] = { a = 1, r = 0.35, g = 0.75, b = 0.25 };
			[COMBATLOG_FILTER_FRIENDLY_UNITS] = { a = 1, r = 0.40, g = 0.50, b = 1.00 };
			[COMBATLOG_FILTER_HOSTILE_UNITS] = { a = 1, r = C.FACTION_BAR_COLORS[2].r, g = C.FACTION_BAR_COLORS[2].g, b = C.FACTION_BAR_COLORS[2].b };
			[COMBATLOG_FILTER_HOSTILE_PLAYERS] = { a = 1, r = C.FACTION_BAR_COLORS[2].r, g = C.FACTION_BAR_COLORS[2].g, b = C.FACTION_BAR_COLORS[2].b };
			[COMBATLOG_FILTER_NEUTRAL_UNITS] = { a = 1, r = C.FACTION_BAR_COLORS[4].r, g = C.FACTION_BAR_COLORS[4].g, b = C.FACTION_BAR_COLORS[4].b };
			[COMBATLOG_FILTER_UNKNOWN_UNITS] = { a = 1, r = 0.75, g = 0.75, b = 0.75 };
		};	
	}

	C.PowerBarColor = setmetatable({
		AMMOSLOT = { r = 0.80, g = 0.60, b = 0.00 };
		ARCANE_CHARGE = { r = 225/255, g = 255/255, b = 255/255 };
		BURNING_EMBERS = { r = 1, g = 0.2, b = 0 };
		DEMONIC_FURY = { r = 222/255, g = 95/255, b = 95/255 };
		ECLIPSE = { negative = { r = 0.0, g = 0.6, b = 1.0 },  positive = { r = 1.0, g = 0.8, b = 0.0 }};
		ENERGY = { r = 1.00, g = 1.00, b = 0.00 };
		FOCUS = { r = 1.00, g = 0.50, b = 0.25 };
		FUEL = { r = 0.00, g = 0.55, b = 0.50 };
		HOLY_POWER = { r = 255/255, g = 225/255, b = 75/255 };
		MANA = { r = 0.00, g = 0.00, b = 1.00 };
		RAGE = { r = 1.00, g = 0.00, b = 0.00 };
		RUNES = { r = 0.50, g = 0.50, b = 0.50 };
		RUNIC_POWER = { r = 0.00, g = 0.82, b = 1.00 };
		SHADOW_ORBS = { r = 0.5, g = 0.5, b = 0.75 };
		SOUL_SHARDS = { r = 0.50, g = 0, b = 1 };
		UNUSED = { r = 0.00; g = 1.00; b = 1.00 };
		
	}, { __index = PowerBarColor })

	-- combopoints for rogues/ferals etc
	C["combopointcolors"] = {
		[1] = { 0.89; 0.00; 0.00 };
		[2] = { 0.89; 0.35; 0.00 };
		[3] = { 0.89; 0.65; 0.00 };
		[4] = { 0.89; 0.89; 0.00 };
		[5] = { 0.00; 0.89; 0.00 };
	}
	
	-- monk chi
	C.Chi = {
		[1] = { 0.69, 0.31, 0.31, 1 };
		[2] = { 0.65, 0.42, 0.31, 1 };
		[3] = { 0.65, 0.63, 0.35, 1 };
		[4] = { 0.46, 0.63, 0.35, 1 };
		[5] = { 0.33, 0.63, 0.33, 1 };
	};

	-- shaman totem colors
	C["totem"] = {
		[EARTH_TOTEM_SLOT] = { r = 0.29; g = 0.56; b = 0.16; }; 
		[FIRE_TOTEM_SLOT] = { r = 0.81; g = 0.39; b = 0.13; }; 
		[WATER_TOTEM_SLOT] = { r = 0.22; g = 0.47; b = 0.75; };
		[AIR_TOTEM_SLOT] = { r = 0.52; g = 0.22; b = 0.90; }; 
	}
end

--------------------------------------------------------------------------------------------------
--		string tags
--------------------------------------------------------------------------------------------------
do
	local bagStrings = {
		["backpack"] = { 0 };
		["bags"] = { 1, 2, 3, 4 };
		["backpack+bags"] = { 0, 1, 2, 3, 4, };
		["bank"] = { 5, 6, 7, 8, 9, 10, 11 };
		["bankframe"] = { -1 };
		["bankframe+bank"] = { -1, 5, 6, 7, 8, 9, 10, 11 };
		["keyring"] = { -2 };
	}

	local getSpace = function(bags)
		if not bags or type(bags) ~= "string" or not bagStrings[bags] then return end

		local free, total, used = 0, 0, 0
		for _,i in pairs(bagStrings[bags]) do
			free, total, used = free + GetContainerNumFreeSlots(i), total + GetContainerNumSlots(i), total - free
		end
		
		return free, total, used, floor(free / total * 100)
	end

	local createIcon = function(iconPath, iconValues)
		if not iconPath or type(iconPath) ~= "string" then return end
		iconValues = type(iconValues) == "string" and iconValues or "0:0:0:0"

		if type(iconValues) == "table" then
			iconValues = tconcat(iconValues, ":")
		end
		
		return ("|T%s:%s|t"):format(iconPath, iconValues)
	end
	
	local goldIcon = createIcon("Interface\\MoneyFrame\\UI-GoldIcon", "0:0:0:0")
	local silverIcon = createIcon("Interface\\MoneyFrame\\UI-SilverIcon", "0:0:0:0")
	local copperIcon = createIcon("Interface\\MoneyFrame\\UI-CopperIcon", "0:0:0:0")

	gUI:RegisterTag("money", function(money, full)
		money = money and tonumber(money) or (GetMoney())

		local str
		local g,s,c = floor(money/1e4), floor(money/100) % 100, money % 100

		if (full) then
			if g > 0 then str = (str and str.."" or "") .. g .. goldIcon end
			if s > 0 or g > 0 then str = (str and str.."" or "") .. s .. silverIcon end
			str = (str and str.."" or "") .. c .. copperIcon
			return str
		else
			if g > 0 then str = (str and str.."" or "") .. g .. goldIcon end
			if s > 0 then str = (str and str.."" or "") .. s .. silverIcon end
			if c > 0 or g + s + c == 0 then str = (str and str.."" or "") .. c .. copperIcon end
			return str
		end
	end)

	gUI:RegisterTag("free", function(bags)
		if not bagStrings[bags] then return end
		return (select(1, getSpace(bags)))
	end)

	gUI:RegisterTag("max", function(bags)
		if not bagStrings[bags] then return end
		return (select(2, getSpace(bags))) 
	end)

	gUI:RegisterTag("used", function(bags)
		if not bagStrings[bags] then return end
		return (select(3, getSpace(bags)))
	end)

	gUI:RegisterTag("freepercent", function(bags)
		if not bagStrings[bags] then return end
		return (select(4, getSpace(bags))).."%"
	end)

	gUI:RegisterTag("usedpercent", function(bags)
		if not bagStrings[bags] then return end
		return (100 - (select(4, getSpace(bags)))).."%"
	end)

	gUI:RegisterTag("shortvalue", function(value)
		value = tonumber(value)

		if not(value) then return "0.0" end
		
		if (value >= 1e6) then
			return ("%.1fM"):format(value / 1e6):gsub("%.?0+([km])$", "%1")
		elseif (value >= 1e3) or (value <= -1e3) then
			return ("%.1fK"):format(value / 1e3):gsub("%.?0+([km])$", "%1")
		else
			return value
		end	
	end)

	-- colored version of the above tag
	gUI:RegisterTag("shortvaluecolored", function(value)
		value = tonumber(value)

		if not(value) then return "0.0" end
		
		if (value >= 1e6) then
			return ("|cFFFFD200%.1f|r|cFFFFFFFFM|r"):format(value / 1e6):gsub("%.?0+([km])$", "%1")
		elseif (value >= 1e3) or (value <= -1e3) then
			return ("|cFFFFD200%.1f|r|cFFFFFFFFK|r"):format(value / 1e3):gsub("%.?0+([km])$", "%1")
		else
			return ("|cFFFFD200%.1f|r"):format(value)
		end	
	end)

	gUI:RegisterTag("smarttime", function(seconds)
		if not(seconds) then return "0.0" end
		
		seconds = tonumber(seconds)
		
		local day, hour, minute = 86400, 3600, 60
		if (seconds >= day) then
			return format("%dd", floor(seconds / day + 0.5)), seconds % day
		elseif (seconds >= hour) then
			return format("%dh", floor(seconds / hour + 0.5)), seconds % hour
		elseif (seconds >= minute) then
			if (seconds <= (minute * 5)) then
				return format("%d:%02d", floor(seconds / 60), seconds % minute), seconds - floor(seconds)
			end
			return format("%dm", floor(seconds / minute + 0.5)), seconds % minute
		elseif (seconds >= minute / 12) then
			return floor(seconds + 0.5), (seconds * 100 - floor(seconds * 100)) / 100
		end
		return format("%.1f", seconds), (seconds * 100 - floor(seconds * 100)) / 100
	end)

	gUI:RegisterTag("time", function(timezone)
		timezone = timezone and timezone:lower() or "game"

		if timezone == "local" or timezone == "local24" then
			return date("%H:%M")

		elseif timezone == "local12" then
			return date("%I:%M %p")

		elseif timezone == "game" or timezone == "game24" then
			local hours, minutes = GetGameTime()
			return (format("%02d", hours)..":"..format("%02d", minutes))

		elseif timezone == "game12" then
			local hours, minutes = GetGameTime()
			return (format("%02d", ((hours > 12) or (hours < 1)) and (hours - 12) or hours)..":"..format("%02d", minutes)..((hours > 12) and " PM" or " AM"))

		elseif timezone == "pvp" then
			local sec = GetPVPTimer()
			if (sec >= 0) and (sec <= 301000) then
				seconds = floor(seconds / 1000)
				return (format("%d", floor(seconds / 60) % 60)..":"..format("%02d", seconds % 60))
			else
				return "--:--"
			end

		elseif timezone == "wintergrasp" then
			local seconds = GetWintergraspWaitTime()
			if seconds then
				return (format("%d", floor(seconds / 3600))..":"..format("%02d", floor(seconds / 60) % 60)..":"..format("%02d", seconds % 60))
			else
				return ("--:--")
			end
		end
	end)
end

--------------------------------------------------------------------------------------------------
--		function library
--------------------------------------------------------------------------------------------------
do
	--------------------------------------------------------------------------------------------------
	--		Character and Area Info
	--------------------------------------------------------------------------------------------------
	do
		local localizedClass, class = UnitClass("player")
		
		-- returns true if the player is in a PvP instance
		F.IsInPvPInstance = function()
			local inInstance, instanceType = IsInInstance()
			return ((inInstance) and ((instanceType == "pvp") or (instanceType == "arena")))
		end

		-- returns true if the player is in an arena
		F.IsInArena = function()
			local inInstance, instanceType = IsInInstance()
			return ((inInstance) and (instanceType == "arena"))
		end

		-- return true if the player is in a World PvP event (Wintergrasp and Tol Barad, and any other they might add)
		F.IsInWorldPvP = function()
			local realZoneName = GetRealZoneText()

			local localizedName, isActive, canQueue, startTime, canEnter
			for pvpID = 1, GetNumWorldPVPAreas() do
				_, localizedName, isActive, canQueue, startTime, canEnter = GetWorldPVPAreaInfo(pvpID)
				if (isActive) and (localizedName == realZoneName) then
					return true
				end
			end
		end

		-- returns true if the player is in a PvP event
		-- this includes Battlegrounds, Arena, Wintergrasp and Tol Barad
		F.IsInPvPEvent = function()
			return (F.IsInPvPInstance()) or (F.IsInWorldPvP())
		end
		
		-- returns true if the player is in any homemade group (not BG made ones)
		F.IsInHomeGroup = function()
			return (GetNumGroupMembers(LE_PARTY_CATEGORY_HOME) > 0)
		end
		
		-- returns true if the player is in a homemade raid (not a BG made one)
		F.IsInHomeRaid = function()
			return (IsInRaid()) and (F.IsInHomeGroup())
		end

		-- returns true if the player is in a homemade party, but NOT a raid
		F.IsInHomeParty = function()
			return not(IsInRaid()) and (F.IsInHomeGroup())
		end

		-- returns true if the player is a raid officer/leader (can set raid marks etc)
		F.IsLeader = function()
			return ((GetNumGroupMembers() > 0) or (GetNumSubgroupMembers() > 0)) and (UnitIsGroupLeader() or UnitIsGroupLeader() or UnitIsGroupAssistant())
		end

		-- returns true if the player is a raid officer/leader, and NOT in a PvP instance or World Event
		F.IsLeaderInPvE = function()
			if (F.IsLeader()) and not(F.IsInPvPEvent()) then
				return true
			end
		end

		-- returns true if the player is a raid officer/leader in a PvP instance or Event
		F.IsLeaderInPvP = function()
			if (F.IsLeader()) and (F.IsInPvPEvent()) then
				return true
			end
		end

		-- only works after the "PLAYER_ALIVE" event!
		F.GetPlayerSpec = function()
			return GetSpecialization() or 0
		end

		-- returns true of the player (or the optional class) is a healerclass
		-- does not check for spec
		F.IsHealerClass = function(classOrNil)
			local class = classOrNil or class
			if (class == "DRUID") or (class == "PALADIN") or (class == "PRIEST") or (class == "SHAMAN") or (class == "MONK") then
				return true
			end
		end

		-- returns true if the player is currently specced as a healer
		F.IsPlayerHealer = function()
			return (GetSpecializationRole(F.GetPlayerSpec()) == "HEALER")
		end

		-- returns true if the player is currently specced as a tank
		F.IsPlayerTank = function()
			return (GetSpecializationRole(F.GetPlayerSpec()) == "TANK")
		end

		-- returns true if the player is currently specced as DPS
		F.IsPlayerDPS = function()
			return (GetSpecializationRole(F.GetPlayerSpec()) == "DAMAGER")
		end
	end

	--------------------------------------------------------------------------------------------------
	--		Taint Handling
	--------------------------------------------------------------------------------------------------
	do
		local safeQueue = {}
		local parseQueue = function(self, event, ...)
			local next
			while (#safeQueue > 0) do
				if (InCombatLockdown()) then
					break
				end
				next = tremove(safeQueue, 1)
				next[1](select(2, next))
			end
		end
		gUI:RegisterEvent("PLAYER_REGEN_ENABLED", parseQueue)
		
		F.SafeCall = function(func, ...)
			if not(InCombatLockdown()) then
				func(...) 
			else
				tinsert(safeQueue, { func, ... })
			end
		end
	end
	
	--------------------------------------------------------------------------------------------------
	--		Version Control
	--------------------------------------------------------------------------------------------------
	do
		--
		-- function used to get the Blizzard build
		-- going to be needing this with all that crap on the 4.3 PTR
		-- @return true if the current WoW version is equal to or greater than the input values
		F.IsBuild = function(version, subversion, tinyversion, build)
			gUI:argCheck(version, 1, "number", "nil")
			gUI:argCheck(subversion, 1, "number", "nil")
			gUI:argCheck(tinyversion, 1, "number", "nil")
			gUI:argCheck(build, 1, "number", "nil")
			local gameversion, gamebuild, gamedate, gametocversion = GetBuildInfo()
			local v, s, t = strsplit(".", gameversion)
				
			if (tonumber(v) >= tonumber(version) or 0) 
			and (tonumber(s) >= (tonumber(subversion) or 0)) 
			and (tonumber(t) >= (tonumber(tinyversion) or 0)) then
				return true
			end
		end
		
		--
		-- returns version, subversion, and Curse build number of the addon if available
		do
			-- clear out letters and clutter, and convert to numbers
			local tbl = {}
			local clean = function(...)
				if not(...) then
					return 0
				end
				wipe(tbl)
				local n
				for i = 1, select("#", ...) do
					n = select(i, ...)
					n = tostring(n) -- make sure it's a string
					n = n:gsub("%D", "") -- remove anything not a number
					n = tonumber(n) -- transform it back to a number
					tinsert(tbl, n or 0) -- don't insert nil values, make them zeroes
				end
				return unpack(tbl)
			end

			-- split a version string
			local version = function(v)
				gUI:argCheck(v, 1, "number", "string", "nil")
				if not(v) then
					return
				else
					v = v:gsub("-", ".") -- consider a minus sign to be the same as a period/dot
					return clean(strsplit(".", v)) -- split it and get the numbers from it
				end
			end

			-- this function is changed from v2 to v3 since we'll be using a different version scheme now
			-- now we will be listing version and subversion in the .toc entry, and build will be from curse
			F.GetVersion = function(test)
				gUI:argCheck(test, 1, "number", "string", "nil")
				local curse = test or GetAddOnMetadata(addon, "X-Curse-Packaged-Version")
				local toc = GetAddOnMetadata(addon, "Version")
				local v,s = version(toc)
				local b = version(curse)
				return v or 0, s or 0, b or 0
			end
		end
	end

	--------------------------------------------------------------------------------------------------
	--		Optionsmenus
	--------------------------------------------------------------------------------------------------
	do
		-- skinned versions of the gOptionsMenu-1.0 widgets
		RegisterMenuWidgets = function()
			------------------------------------------------------------------------------------------------------------
			-- 	Widgets
			------------------------------------------------------------------------------------------------------------
			local containerWidth, containerHeight = 623, 568
			local SetTooltipScripts = function(self, hook)
				local SetScript = hook and "HookScript" or "SetScript"
				
				self[SetScript](self, "OnEnter", function(self)
					if (self.tooltipText) then
						GameTooltip:SetOwner(self, self.tooltipOwnerPoint or "ANCHOR_RIGHT")

						if (type(self.tooltipText) == "string") then
							GameTooltip:SetText(self.tooltipText, 1.0, .82, .0, 1.0, 1)
							
						elseif (type(self.tooltipText) == "table") then
							for i = 1, #self.tooltipText do
								if (i == 1) then
									GameTooltip:SetText(self.tooltipText[i], 1.0, 1.0, 1.0, 1.0, 1)
								else
									GameTooltip:AddLine(self.tooltipText[i], 1.0, .82, .0, 1.0)
								end
							end
						end
						
						if (self.tooltipRequirement) then
							GameTooltip:AddLine(self.tooltipRequirement, 1.0, .0, .0, 1.0)
						end

						GameTooltip:Show()
					end
				end)
				
				self[SetScript](self, "OnLeave", function(self)
					GameTooltip:Hide()
				end)
			end
			
			-- make sure all objects are represented here, or it will bug out
			local ObjectPadding = {
				-- text objects
				Title = { x = { before = 0, after = 0 }, y = { before = 16, after = 16 } };
				Header = { x = { before = 0, after = 0 }, y = { before = 16, after = 8 } };
				Text = { x = { before = 0, after = 0 }, y = { before = 0, after = 8 } };
				
				-- image objects
				Texture = { x = { before = 0, after = 0 }, y = { before = 0, after = 0 } };
				
				-- frame groups
				Frame = { x = { before = 0, after = 0 }, y = { before = 0, after = 0 } };
				ScrollFrame = { x = { before = 0, after = 0 }, y = { before = 0, after = 0 } };

				-- input widgets
				Button = { x = { before = 8, after = 8 }, y = { before = 8, after = 8 } };
				CheckButton = { x = { before = 8, after = 0 }, y = { before = 4, after = 4 } };
				ColorSelect = { x = { before = 0, after = 0 }, y = { before = 0, after = 0 } };
				Dropdown = { x = { before = 8, after = 0 }, y = { before = 0, after = 0 } };
				EditBox = { x = { before = 0, after = 0 }, y = { before = 0, after = 0 } };
				RadioButton = { x = { before = 8, after = 0 }, y = { before = 4, after = 4 } };
				Slider = { x = { before = 16, after = 16 }, y = { before = 16, after = 16 } };
				StatusBar = { x = { before = 16, after = 16 }, y = { before = 8, after = 8 } };
				TabButton = { x = { before = 0, after = 0 }, y = { before = 0, after = 0 } };
			}
			local Widgets = {
				-- use for page titles
				Title = function(parent, msg, name, ...)
					local self = parent:CreateFontString(name, "ARTWORK")
					self.isEnabled = true
					self:SetFontObject(gUI_TextFontLarge)
					self:SetTextColor(unpack(C["value"]))
					self:SetJustifyH("LEFT")
					self:SetWordWrap(true)
					self:SetNonSpaceWrap(true)
					self:SetText(msg)

					self.Enable = function(self) 
						self.isEnabled = true
						self:SetTextColor(unpack(C["value"]))
					end

					self.Disable = function(self) 
						self.isEnabled = false
						self:SetTextColor(unpack(C["disabled"]))
					end

					self.IsEnabled = function(self) return self.isEnabled end
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end
					
					return self, ...
				end;

				-- use for paragraph headers
				Header = function(parent, msg, name, ...)
					local self = parent:CreateFontString(name, "ARTWORK")
					self.isEnabled = true
					self:SetFontObject(gUI_TextFontNormal)
					self:SetTextColor(unpack(C["index"]))
					self:SetJustifyH("LEFT")
					self:SetWordWrap(true)
					self:SetNonSpaceWrap(true)
					self:SetText(msg)
					
					self.Enable = function(self) 
						self.isEnabled = true
						self:SetTextColor(unpack(C["index"]))
					end

					self.Disable = function(self) 
						self.isEnabled = false
						self:SetTextColor(unpack(C["disabled"]))
					end

					self.IsEnabled = function(self) return self.isEnabled end

					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("LEFT", parent, "RIGHT", 8, 0)
					end

					return self, ...
				end;

				-- use for normal text
				Text = function(parent, msg, name, ...)
					local self = parent:CreateFontString(name, "ARTWORK")
					self.isEnabled = true
					self:SetFontObject(gUI_TextFontSmallWhite)
					self:SetTextColor(unpack(C["index"]))
					self:SetJustifyH("LEFT")
					self:SetWordWrap(true)
					self:SetNonSpaceWrap(true)
					self:SetText((type(msg) == "table") and tconcat(msg, "|n") or msg)
					
					self.Enable = function(self) 
						self.isEnabled = true
						self:SetTextColor(unpack(C["index"]))
					end

					self.Disable = function(self) 
						self.isEnabled = false
						self:SetTextColor(unpack(C["disabled"]))
					end

					self.IsEnabled = function(self) return self.isEnabled end
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("LEFT", parent, "RIGHT", 8, 0)
					end

					return self, ...
				end;
				
				Texture = function(parent, msg, name, ...)
					local self = parent:CreateTexture(name, "ARTWORK")
					self:SetSize(32, 32)
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("LEFT", parent, "RIGHT", 8, 0)
					end

					return self, ...
				end;

				Button = function(parent, msg, name, ...)
					local self = CreateFrame("Button", name, parent, "UIPanelButtonTemplate")
					self:SetSize(80, 22)
					self:SetText(msg)
					gUI:SetUITemplate(self, "button", true)
					
					self:SetScript("OnClick", function(self)
						if (self.set) then
							self:set()
							
						elseif (self.parent.set) then
							self.parent:set()
						end
					end)

					SetTooltipScripts(self)
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end
					
					return self, ...
				end;

				CheckButton = function(parent, msg, name, ...)
					local self = CreateFrame("CheckButton", name, parent, "OptionsBaseCheckButtonTemplate")  -- OptionsBaseCheckButtonTemplate?
					
					local text = self:CreateFontString(name .. "Text", "ARTWORK")
					text:SetFontObject(gUI_TextFontSmallWhite)
					text:SetPoint("LEFT", self, "RIGHT", 8, 0)
					text:SetTextColor(unpack(C["index"]))
					text:SetWordWrap(true)
					text:SetNonSpaceWrap(true)
					text:SetText(msg)
					self.text = text	
					gUI:SetUITemplate(self, "checkbutton"):SetBackdropColor(nil, nil, nil, gUI:GetOverlayAlpha())

					self.refresh = function(self, option)
						if (self.get) then
							self:SetChecked(option or self:get())
							
						elseif (self.parent.get) then
							self:SetChecked(option or self.parent:get())
						end

						if (self.onrefresh) then
							self:onrefresh()
						end
					end

					self:SetScript("OnShow", function(self) self:refresh() end)
					self:SetScript("OnEnable", function(self) self.text:SetTextColor(unpack(C["index"])) end)
					self:SetScript("OnDisable", function(self) self.text:SetTextColor(unpack(C["disabled"])) end)
					self:SetScript("OnClick", function(self)
						if (self:GetChecked()) then
							PlaySound("igMainMenuOptionCheckBoxOn")
						else
							PlaySound("igMainMenuOptionCheckBoxOff")
						end
						
						if (self.set) then
							self:set()
							
						elseif (self.parent.set) then
							self.parent:set()
						end
					end)
					
					SetTooltipScripts(self)
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self, ...
				end;

				Frame = function(parent, name, ...)
					local self = CreateFrame("Frame", name, parent or UIParent)
					self:SetSize(containerWidth, containerHeight)
					self:EnableMouse(false)
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self
				end;

				ScrollFrame = function(parent, name, ...)
					local self = CreateFrame("ScrollFrame", name, parent or UIParent) -- "UIPanelScrollFrameTemplate"
					self:SetSize(containerWidth - 32 - 16, containerHeight - 32)
					self:EnableMouseWheel(true)
					
					self.ScrollChild = CreateFrame("Frame", name .. "ScrollChild", self)
					self.ScrollChild:SetSize(self:GetSize())
					self.ScrollChild:SetAllPoints(self)
					
					self:SetScrollChild(self.ScrollChild)
					self:SetVerticalScroll(0)
					
					self.ScrollBar = CreateFrame("Slider", name .. "ScrollBar", self, "UIPanelScrollBarTemplate")
					self.ScrollBar:SetPoint("TOPLEFT", self, "TOPRIGHT", 6, -16)
					self.ScrollBar:SetPoint("BOTTOMLEFT", self, "BOTTOMRIGHT", 6, 16)
					self.ScrollBar:SetWidth(16)
					self.ScrollBar:SetMinMaxValues(0, 0)
					self.ScrollBar:SetValue(0)

					self.ScrollBar.up = _G[name .. "ScrollBarScrollUpButton"]
					self.ScrollBar.up:Disable()
					self.ScrollBar.up:SetScript("OnClick", function(self)
						local ScrollBar = self:GetParent()
						local ScrollFrame = self:GetParent():GetParent()
						local scrollStep = ScrollFrame.scrollStep or (ScrollBar:GetHeight() / 3)

						ScrollBar:SetValue(min(0, ScrollBar:GetValue() - scrollStep))
						
						PlaySound("UChatScrollButton")
					end)
					
					self.ScrollBar.down = _G[name .. "ScrollBarScrollDownButton"]
					self.ScrollBar.down:Disable()
					self.ScrollBar.down:SetScript("OnClick", function(self)
						local ScrollBar = self:GetParent()
						local ScrollFrame = self:GetParent():GetParent()
						local scrollStep = ScrollFrame.scrollStep or (ScrollFrame:GetHeight() / 3)

						ScrollBar:SetValue(min(ScrollFrame:GetVerticalScrollRange(), ScrollBar:GetValue() + scrollStep))

						PlaySound("UChatScrollButton")
					end)
					gUI:SetUITemplate(self.ScrollBar, "scrollbar")
					
					self.Update = function(self)
						local w, h = self:GetSize()
						local sW, sH = self.ScrollChild:GetSize()

						if (w ~= sW) then
							self.ScrollChild:SetWidth(w)
						end

						if (h ~= sH) then
							self.ScrollChild:SetHeight(h)
						end
						
						self:UpdateScrollChildRect()

						local min, max, value = 0, self:GetVerticalScrollRange(), self:GetVerticalScroll()
						
						if (value > max) then
							value = max
						end
						
						if (value < min) then
							value = min
						end
						
						self.ScrollBar:SetMinMaxValues(min, max)
						
						if (value == min) then
							if (self.ScrollBar.up:IsEnabled()) then
								self.ScrollBar.up:Disable()
							end

							if not(self.ScrollBar.down:IsEnabled()) then
								self.ScrollBar.down:Enable()
							end
							
						elseif (value == max) then
							if (self.ScrollBar.down:IsEnabled()) then
								self.ScrollBar.down:Disable()
							end
							
							if not(self.ScrollBar.up:IsEnabled()) then
								self.ScrollBar.up:Enable()
							end
						else
							if not(self.ScrollBar.up:IsEnabled()) then
								self.ScrollBar.up:Enable()
							end

							if not(self.ScrollBar.down:IsEnabled()) then
								self.ScrollBar.down:Enable()
							end
						end
					end

					self.ScrollBar:SetScript("OnValueChanged", function(self, value)
						self:GetParent():SetVerticalScroll(value)
						self:GetParent():Update()
					end)
					
					self:SetScript("OnMouseWheel", function(self, delta)
						if (delta > 0) then
							if (self.ScrollBar.up:IsEnabled()) then
								self.ScrollBar:SetValue(max(0, self.ScrollBar:GetValue() - 20))
							end
							
						elseif (delta < 0) then
							if (self.ScrollBar.down:IsEnabled()) then
								self.ScrollBar:SetValue(min(self:GetVerticalScrollRange(), self.ScrollBar:GetValue() + 20))
							end
						end
					end)
					
					-- we schedule a timer to update the frame contents 1/5 second after it's shown
					-- we only do this the first time
					local once
					self:SetScript("OnShow", function(self) 
						if not(once) then
							gUI:ScheduleTimer(1/5, function() self:Update() end)
							once = true
						end
					end)
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self, self:GetScrollChild()
				end;
				
				ColorSelect = function(parent, msg, name, ...)
					local self = CreateFrame("ColorSelect", name, parent)

					SetTooltipScripts(self)
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self, ...
				end;

				Dropdown = function(parent, msg, name, args, width, ...)
					local width = width or 100
					local self = CreateFrame("Button", name, parent, "UIDropDownMenuTemplate")
					self:SetHitRectInsets(-26, 0, 0, 0)
					gUI:SetUITemplate(self, "dropdown", true)
					
					SetTooltipScripts(self)
					
					local label = self:CreateFontString(name .. "Label", "ARTWORK")
					label:SetFontObject(gUI_TextFontSmallWhite)
					label:SetPoint("LEFT", self, "RIGHT", 0, 0)
					label:SetText(msg)
					self.label = label
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end
					
					local onclick = function(self)
						-- select the item you clicked on
						UIDropDownMenu_SetSelectedID(_G[name], self:GetID())

						-- fire off the button's 'set' function, and pass the ID along
						_G[name].set(_G[name], self:GetID())
						_G[name].selectedID = self:GetID()
					end
					
					self.args = CopyTable(args)
					self.refresh = function(self, option)
						if (self.get) then
							self.selectedID = self:get()
						end
						
						option = option or self.selectedID
						
						if (option) and (self.args[option]) then
							_G[name .. "Text"]:SetText(self.args[option])
						end

						if (self.onrefresh) then
							self:onrefresh()
						end
					end
					
					self:HookScript("OnShow", function(self) self:refresh() end)
			--		self:HookScript("OnHide", function(self) self:refresh() end)
					
					local info = {}
					local init = function(self, level)
			--			for i,v in pairs(args) do
						for i = 1, #args do
							wipe(info)
					
							info = UIDropDownMenu_CreateInfo()
							info.text = args[i] -- v
							info.value = i
							info.func = onclick

							UIDropDownMenu_AddButton(info, level)
						end
					end

					UIDropDownMenu_Initialize(self, init)
					UIDropDownMenu_SetWidth(self, width)
					UIDropDownMenu_SetButtonWidth(self, width)
					UIDropDownMenu_JustifyText(self, "LEFT")
					UIDropDownMenu_SetSelectedID(self, 1) -- selecting option #1 as default

					return self, ...
				end;

				EditBox = function(parent, msg, name, args, ...)
					local self = CreateFrame("Frame", nil, parent)
					self:EnableMouse(true)
					self:SetSize(90, 14)
					self:SetScript("OnMouseDown", function(self) self.editBox:Show() end)

					local text = self:CreateFontString(name .. "Text", "ARTWORK")
					text:SetFontObject(gUI_DisplayFontSmall)
					text:SetPoint("BOTTOMLEFT", 0, 2)
					text:SetJustifyH("LEFT")
					text:SetJustifyV("BOTTOM")
					text:SetText("")
					text:SetTextColor(unpack(C["value"]))
					self.text = text
					
					local suffix = self:CreateFontString(name .. "TextSuffix", "ARTWORK")
					suffix:SetFontObject(gUI_TextFontSmallWhite)
					suffix:SetPoint("BOTTOMLEFT", text, "BOTTOMRIGHT")
					suffix:SetJustifyH("LEFT")
					suffix:SetJustifyV("BOTTOM")
					suffix:SetTextColor(unpack(C["index"]))
					suffix:SetText(msg)
					self.suffix = suffix
					
					local editBox = CreateFrame("EditBox", nil, self)
					editBox.parent = self
					editBox:Hide()
					editBox:SetSize(self:GetWidth() + 8, self:GetHeight() + 8)
					editBox:SetPoint("BOTTOMLEFT", -4, -2)
					editBox:SetJustifyH("LEFT")
					editBox:SetJustifyV("BOTTOM")
					editBox:SetTextInsets(4, 4, 0, 0)
					editBox:SetFontObject(gUI_DisplayFontSmall)
					editBox:SetAutoFocus(false)
					editBox:SetNumeric((args) and args.numeric)
					
					gUI:SetUITemplate(self, "editbox")
					
					editBox.Refresh = function(self) 
						if (self.parent.get) then
							if (self:IsNumeric()) then
								self:SetNumber(self.parent:get())
							else
								self:SetText(self.parent:get())
							end
						else
							if (self:IsNumeric()) then
								self:SetNumber("")
							else
								self:SetText("")
							end
						end
					end

					editBox:SetScript("OnHide", function(self) 
						self.parent.text:Show()
						self.parent.suffix:Show()
					end)
					
					editBox:SetScript("OnShow", function(self) 
						self.parent.text:Hide()
						self.parent.suffix:Hide()
						
						self:Refresh()
						self:SetFocus()
						self:HighlightText()
					end)
					
					editBox:SetScript("OnEditFocusLost", editBox.Hide)
					editBox:SetScript("OnEscapePressed", editBox.Hide)
					editBox:SetScript("OnEnterPressed", function(self) 
						self:Hide()
						
						local msg = self:IsNumeric() and self:GetNumber() or self:GetText()
						if (msg) then
							if (self.parent.set) then
								self.parent:set(msg)
							end
						end
						
						self.parent:refresh()
					end)
					
					self.editBox = editBox
					
					SetTooltipScripts(self)
					
					self.refresh = function(self)
						if (self.get) then
							self.text:SetText(self.get())
						else
							self.text:SetText("")
						end
						
						if (self.editBox:IsShown()) then
							self.editBox:Refresh()
						end
						
						if (self.onrefresh) then
							self:onrefresh()
						end
					end
					
					self:HookScript("OnSizeChanged", function(self) 
						self.editBox:SetSize(self:GetWidth() + 8, self:GetHeight() + 8)
					end)
					
					self.Enable = function(self) 
						self.isEnabled = true
						self:EnableMouse(true)
						self.text:SetTextColor(unpack(C["value"]))
						self.suffix:SetTextColor(unpack(C["index"]))
					end

					self.Disable = function(self) 
						self.isEnabled = false
						self:EnableMouse(false)
						self.text:SetTextColor(unpack(C["disabled"]))
						self.suffix:SetTextColor(unpack(C["disabled"]))
						if (self.editBox:IsShown()) then
							self.editBox:Hide()
						end
					end

					self.IsEnabled = function(self) return self.isEnabled end
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self, ...
				end;

				RadioButton = function(parent, msg, name, ...)
					local self = CreateFrame("CheckButton", name, parent, "UIRadioButtonTemplate")

					local text = self:CreateFontString(name .. "Text", "ARTWORK")
					text:SetFontObject(gUI_TextFontSmallWhite)
					text:SetPoint("LEFT", self, "RIGHT", 8, 0)
					text:SetTextColor(unpack(C["index"]))
					text:SetWordWrap(true)
					text:SetNonSpaceWrap(true)
					text:SetText(msg)
					self.text = text	
					gUI:SetUITemplate(self, "radiobutton"):SetBackdropColor(nil, nil, nil, gUI:GetOverlayAlpha())

					self.refresh = function(self, option)
						if (self.get) then
							self:SetChecked(option or self:get())
							
						elseif (self.parent.get) then
							self:SetChecked(option or self.parent:get())
						end

						if (self.onrefresh) then
							self:onrefresh()
						end
					end

					self:SetScript("OnShow", function(self) self:refresh() end)
					self:SetScript("OnEnable", function(self) self.text:SetTextColor(unpack(C["index"])) end)
					self:SetScript("OnDisable", function(self) self.text:SetTextColor(unpack(C["disabled"])) end)
					self:SetScript("OnClick", function(self)
						if (self:GetChecked()) then
							PlaySound("igMainMenuOptionCheckBoxOn")
						else
							PlaySound("igMainMenuOptionCheckBoxOff")
						end
						
						if (self.set) then
							self:set()
							
						elseif (self.parent.set) then
							self.parent:set()
						end
					end)

					SetTooltipScripts(self)
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self, ...
				end;

				Slider = function(parent, msg, name, orientation, ...)
					orientation = orientation or "HORIZONTAL"
					
					local self = CreateFrame("Slider", name, parent, "OptionsSliderTemplate")
					self:SetOrientation(orientation)
					gUI:SetUITemplate(self, "slider")
					
					self.low = _G[name .. "Low"]
					self.low:SetFontObject(gUI_DisplayFontExtraTiny)

					self.high = _G[name .. "High"]
					self.high:SetFontObject(gUI_DisplayFontExtraTiny)

					self.text = _G[name .. "Text"]
					self.text:SetFontObject(gUI_DisplayFontTiny)
					
					self.refresh = function(self, option)
						if (self.get) then
							local value = self:get()
							if (value) then
								self:SetValue(value)
							end
						end
					end

					self:SetScript("OnShow", function(self) self:refresh() end)
					
					self:SetScript("OnValueChanged", function(self, value)
						if (self.set) then
							self:set(value)
						end
					end)
					
					SetTooltipScripts(self)

					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self, ...
				end;

				StatusBar = function(parent, msg, name, ...)
					local self = CreateFrame("StatusBar", name, parent)
					gUI:SetUITemplate(self, "statusbar", true)
					
					SetTooltipScripts(self)
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self, ...
				end;
				
				TabButton = function(parent, msg, name, ...)
					local self = CreateFrame("CheckButton", name, parent, "TabButtonTemplate")
					gUI:SetUITemplate(self, "tab")
					
					if (...) then
						self:SetPoint(...)
					else
						self:SetPoint("TOPLEFT")
					end

					return self, ...
				end;
			}
			
			local padding
			for name, func in pairs(Widgets) do
				padding = ObjectPadding[name]
				gUI:RegisterOptionsMenuWidgetClass(name, func, padding.x.before, padding.x.after, padding.y.before, padding.y.after)
			end
		end
	
		-- game menu buttons
		do
			local menuButtons = {}
			local topButton 
			F.AddGameMenuButton = function(name, msg, onclick)
				-- create our menu button
				local button = CreateFrame("Button", "GameMenuButton" .. name or "", GameMenuFrame, "GameMenuButtonTemplate")
				button:SetText(msg)
				button:SetScript("OnClick", onclick)

				-- find the top blizzard button
				-- this should be both backwards and forward compatible, and quite possible work with other addons as well
				if not(topButton) then
					for i = 1, GameMenuFrame:GetNumChildren() do
						local child = select(i, GameMenuFrame:GetChildren())
						local _, b, c, _, _ = child:GetPoint()
						if (b == GameMenuFrame) and (c == "TOP") then
							topButton = child
							break
						end
					end
				end
				
				-- decide where exactly to put our new button
				if (#menuButtons == 0) then
					-- take blizzard's top button's place
					button:SetPoint(topButton:GetPoint())
				else
					-- but it below the last of our custom buttons
					button:SetPoint("TOP", menuButtons[#menuButtons], "BOTTOM", 0, 0)
				end
				
				-- move the rest of the menu buttons down
				-- they will always come 16 points under our last custom button
				topButton:ClearAllPoints()
				topButton:SetPoint("TOP", button, "BOTTOM", 0, -16)
				
				-- make a global reference to our button
				_G["GameMenuButton" .. name or ""] = button
				
				-- increase the size of the GameMenuFrame to make place for our new button
				GameMenuFrame:SetHeight(GameMenuFrame:GetHeight() + button:GetHeight() + ((#menuButtons == 0) and 16 or 0))

				-- add it to our list
				tinsert(menuButtons, button)
			end

			F.GetGameMenuButtons = function(self)
				return menuButtons
			end
		end

		-- restart scheduling
		do
			local warned, scheduledRestart
			
			F.red = function(msg)
				gUI:argCheck(msg, 1, "string")
				return "|cFFFF0000" .. msg .. "|r"
			end

			F.green = function(msg)
				gUI:argCheck(msg, 1, "string")
				return "|cFF00FF00" .. msg .. "|r"
			end

			F.yellow = function(msg)
				gUI:argCheck(msg, 1, "string")
				return "|cFFFFD100" .. msg .. "|r"
			end

			-- make text red. very advanced function. oh yeah.
			F.warning = function(...)
				return F.red(...)
			end

			-- schedule a restart, without actually doing anything
			F.ScheduleRestart = function()
				scheduledRestart = true
			end
			
			-- cancel the scheduled restart
			F.CancelRestart = function()
				scheduledRestart = nil
				print(L["You can reload the user interface at any time with |cFF4488FF/rl|r or |cFF4488FF/reload|r"])
			end

			-- popup a restart request if one has been scheduled previously
			F.RestartIfScheduled = function()
				if (scheduledRestart) and not(warned) then
					StaticPopup_Show("GUIS_RESTART_REQUIRED_FOR_CHANGES")
					warned = true
				end
			end
		end
	end
	
	--------------------------------------------------------------------------------------------------
	--		Styling
	--------------------------------------------------------------------------------------------------
	do
		-- http://www.wowpedia.com/ColorGradient
		local ColorGradient = function(a, b, ...)
			local perc
			if(b == 0) then
				perc = 0
			else
				perc = a / b
			end

			if perc >= 1 then
				local r, g, b = select(select('#', ...) - 2, ...)
				return r, g, b
			elseif perc <= 0 then
				local r, g, b = ...
				return r, g, b
			end

			local num = select('#', ...) / 3
			local segment, relperc = math.modf(perc*(num-1))
			local r1, g1, b1, r2, g2, b2 = select((segment*3)+1, ...)

			return r1 + (r2-r1)*relperc, g1 + (g2-g1)*relperc, b1 + (b2-b1)*relperc
		end
		
		-- proxy function that uses a basic red-yellow-green gradient if no other is specified
		F.ColorGradient = function(a, b, ...)
			if (...) then
				return ColorGradient(a, b, ...)
			else
				return ColorGradient(a, b, 1, 0, 0, 1, 1, 0, 0, 1, 0)
			end
		end
	
		F.GetDurabilityColor = function(current, maximum)
			gUI:argCheck(current, 1, "number", "nil")
			gUI:argCheck(maximum, 1, "number", "nil")
			local r, g, b = 1, 1, 1
			if ((maximum) and (maximum > 0)) then
				r, g, b = unpack(C["durability"][(current or 0)/maximum * 100])
			end
			return r, g, b
		end

		F.GetRarityText = function(rarity)
			gUI:argCheck(rarity, 1, "number")
			local r, g, b, hex = GetItemQualityColor(rarity)
			return "|c" .. hex .. _G["ITEM_QUALITY" .. rarity .. "_DESC"] .. "|r"
		end
	end

	--------------------------------------------------------------------------------------------------
	--		Shine
	--------------------------------------------------------------------------------------------------
	--
	-- usage:
	-- 	local shineFrame = F.Shine:New(target, maxAlpha, duration, scale)
	-- 	shineFrame:Start()
	--		shineFrame:Hide()
	do
		local MAXALPHA = 1
		local SCALE = 5 -- too huge for this?
		local DURATION = 0.75
		local TEXTURE = M("Texture", "CooldownStar") -- [[Interface\Cooldown\star4]]

		local New = function(frameType, parentClass)
			gUI:argCheck(frameType, 1, "string")
			gUI:argCheck(parentClass, 2, "table", "nil")
			local class = CreateFrame(frameType)
			class.mt = { __index = class }
			if (parentClass) then
				class = setmetatable(class, { __index = parentClass })
				class.super = function(self, method, ...)
					parentClass[method](self, ...)
				end
			end
			class.Bind = function(self, obj) return setmetatable(obj, self.mt) end
			return class
		end

		local Shine = New("Frame")

		Shine.New = function(self, parent, maxAlpha, duration, scale)
			local f = self:Bind(CreateFrame("Frame", nil, parent)) 
			f:Hide()
			f:SetScript("OnHide", f.OnHide)
			f:SetAllPoints(parent)
			f:SetToplevel(true)

			f.animation = f:CreateShineAnimation(maxAlpha, duration, scale)

			local icon = f:CreateTexture(nil, "OVERLAY")
			icon:SetPoint("CENTER")
			icon:SetBlendMode("ADD")
			icon:SetAllPoints(f)
			icon:SetTexture(TEXTURE)

			return f
		end

		local animation_OnFinished = function(self)
			local parent = self:GetParent()
			if (parent:IsShown()) then
				parent:Hide()
			end
		end

		Shine.CreateShineAnimation = function(self, maxAlpha, duration, scale)
			local MAXALPHA = maxAlpha or MAXALPHA
			local SCALE = scale or SCALE
			local DURATION = duration or DURATION

			local g = self:CreateAnimationGroup()
			g:SetLooping("NONE")
			g:SetScript("OnFinished", animation_OnFinished)

			local startTrans = g:CreateAnimation("Alpha")
			startTrans:SetChange(-1) -- make it 0, no matter the maxAlpha
			startTrans:SetDuration(0)
			startTrans:SetOrder(0)

			local grow = g:CreateAnimation("Scale")
			grow:SetOrigin("CENTER", 0, 0)
			grow:SetScale(SCALE, SCALE)
			grow:SetDuration(DURATION/2)
			grow:SetOrder(1)

			local brighten = g:CreateAnimation("Alpha")
			brighten:SetChange(MAXALPHA)
			brighten:SetDuration(DURATION/2)
			brighten:SetOrder(1)

			local shrink = g:CreateAnimation("Scale")
			shrink:SetOrigin("CENTER", 0, 0)
			shrink:SetScale(-SCALE, -SCALE)
			shrink:SetDuration(DURATION/2)
			shrink:SetOrder(2)

			local fade = g:CreateAnimation("Alpha")
			fade:SetChange(-MAXALPHA)
			fade:SetDuration(DURATION/2)
			fade:SetOrder(2)

			return g
		end

		Shine.OnHide = function(self)
			if (self.animation:IsPlaying()) then
				self.animation:Finish()
			end
			self:Hide()
		end

		Shine.Start = function(self)
			if (self.animation:IsPlaying()) then
				self.animation:Finish()
			end
			self:Show()
			self.animation:Play()
		end
		
		-- global access
		F.Shine = Shine
	end
	
end

--------------------------------------------------------------------------------------------------
--		Skinner API (moved to the core in v3)
--------------------------------------------------------------------------------------------------
do
	-- removal
	do
		-- need some local debugging in my conversion process from v2 here
		local prevName
		local textures = {}
		local tex = UIParent:CreateTexture() -- just for... stuff
		gUI.HideTexture = function(self, texture)
			if not(texture) then
				print("HideTexture: The object after " .. (prevName or "nil") .. " doesn't exist!")
				return
			else
				prevName = texture:GetName()
			end
			if not(textures[texture]) then
				textures[texture] = texture:GetTexture()
				texture:SetTexture("")
				texture.SetTexture = self.noop
			end
		end

		gUI.ShowTexture = function(self, texture)
			if (textures[texture]) then
				texture.SetTexture = tex.SetTexture
				texture:SetTexture(textures[texture])
				textures[texture] = nil
			end
		end

		gUI.DisableTextures = function(self, object)
			if not(object) then
				print("DisableTextures: The object after " .. (prevName or "nil") .. " doesn't exist!")
				return
			else
				prevName = object:GetName()
			end
			if (object:GetObjectType() == "Texture") then
				print("DisableTextures: The object " .. prevName .. " is a texture, use gUI:HideTexture() instead")
				return
			end
			local region
			for i = 1, object:GetNumRegions() do
				region = select(i, object:GetRegions())
				if (region:GetObjectType() == "Texture") then
					self:HideTexture(region)
				end
			end
		end

		gUI.EnableTextures = function(self, object)
			local region
			for i = 1, object:GetNumRegions() do
				region = select(i, object:GetRegions())
				if (region:GetObjectType() == "Texture") then
					self:ShowTexture(region)
				end
			end
		end

		local clutter = { "Left"; "Right"; "Middle"; "Mid"; "Top"; "Bottom"; "Background"; "BG"; "Track"; "TopRight"; "TopLeft"; "BottomLeft"; "BottomRight" }
		gUI.RemoveClutter = function(self, object)
			if not(object) or not(object.GetName) or not(object:GetName()) then
				return
			end
			
			local t
			for i,v in pairs(clutter) do
				t = _G[object:GetName() .. v]
				if (t) then
					self:KillObject(t)
				end
			end
		end
		
		local frames = {}
		gUI.HideFrame = function(self, frame)
			if not(frames[frame]) then
				frames[frame] = frame:GetParent()
				frame:SetParent(self:GetAttribute("dummy"))
			end
		end
		
		gUI.ShowFrame = function(self, frame)
			if (frames[frame]) then
				frame:SetParent(frames[frame])
				frames[frame] = nil
			end
		end
		
		gUI.HideObject = function(self, object)
			local type = object:GetObjectType()
			local isFrame = object:IsObjectType("Frame")
			if (type == "Texture") then
				self:HideTexture(object)
			elseif (isFrame) then
				self:HideFrame(object)
			end
		end
		
		gUI.HideObjects = function(self, ...)
			for i = 1, select("#", ...) do
				self:HideObject(select(i, ...))
			end
		end

		gUI.ShowObject = function(self, object)
			local type = object:GetObjectType()
			local isFrame = object:IsObjectType("Frame")
			if (type == "Texture") then
				self:ShowTexture(object)
			elseif (isFrame) then
				self:ShowFrame(object)
			end
		end
		
		gUI.ShowObjects = function(self, ...)
			for i = 1, select("#", ...) do
				self:ShowObject(select(i, ...))
			end
		end

		gUI.KillObject = function(self, object)
			if not(object) then
				print("KillObject: The object after " .. (prevName or "nil") .. " doesn't exist!")
				return
			else
				prevName = object:GetName()
			end
			if (object.RegisterEvent) then
				object:UnregisterAllEvents()
			end
			self:HideFrame(object)
		end
		
		-- change a cvar, but only if it's not already the value we want
		-- we do this to avoid firing off hooks related to SetCVar when not needed
		gUI.SetCVar = function(self, cvar, value)
			local c
			if (type(value) == "number") then
				c = tonumber(GetCVar(cvar))
			else
				c = GetCVar(cvar)
			end
			if (c ~= value) then
				_G.SetCVar(cvar, value)
			end
		end
		
		-- remove an entire blizzard options panel, 
		-- and disable its automatic cancel/okay functionality
		-- this is needed, or the option will be reset when the menu closes
		-- it is also a major source of taint related to the Compact group frames!
		gUI.KillPanel = function(self, i, panel)
			if (i) then
				local cat = _G["InterfaceOptionsFrameCategoriesButton" .. i]
				if (cat) then
					cat:SetScale(0.00001)
					cat:SetAlpha(0)
				end
			end
			if (panel) then
				self:KillObject(panel)
				panel.cancel = self.noop
				panel.okay = self.noop
				panel.refresh = self.noop
			end
		end
		
		-- remove a blizzard menu option, 
		-- and disable its automatic cancel/okay functionality
		-- this is needed, or the option will be reset when the menu closes
		-- it is also a major source of taint related to the Compact group frames!
		gUI.KillOption = function(self, shrink, option)
			if not(option) or not(option.IsObjectType) or not(option:IsObjectType("Frame")) then
				return
			end
			self:KillObject(option)
			if (shrink) then
				option:SetHeight(0.00001)
			end
			option.cvar = ""
			option.uvar = ""
			option.value = nil
			option.oldValue = nil
			option.defaultValue = nil
			option.setFunc = self.noop
		end
		
	end
	
	-- styling
	do
		local makeBackdropFrame = function(object, inset, level)
			inset = inset or 0
			level = level or 5
			local oldLevel = object:GetFrameLevel()
			local backdropFrame = CreateFrame("Frame", nil, object)
			backdropFrame:SetPoint("TOPLEFT", object, "TOPLEFT", -inset, inset)
			backdropFrame:SetPoint("BOTTOMRIGHT", object, "BOTTOMRIGHT", inset, -inset)
			backdropFrame:SetFrameLevel(max(oldLevel - level, 0))
			hooksecurefunc(object, "SetFrameLevel", function(object, newLevel) 
				backdropFrame:SetFrameLevel(max(newLevel - level, 0))
			end)
			return backdropFrame
		end
		
		local SetHoverScripts = function(object)
			if not(object._hasHoverScripts) then
				object.__originalRGB = { object:GetBackdropColor() }
				object.__originalBorderRGB = { object:GetBackdropBorderColor() }
				
				object:HookScript("OnEnter", function(object) 
		--			object:SetBackdropColor(C["hovercolor"][1], C["hovercolor"][2], C["hovercolor"][3])
					object:SetBackdropColor(0, 0, 0)
					object:SetBackdropBorderColor(C["hoverbordercolor"][1], C["hoverbordercolor"][2], C["hoverbordercolor"][3])
				end)

				object:HookScript("OnLeave", function(object)
					object:SetBackdropColor(unpack(object.__originalRGB))
					object:SetBackdropBorderColor(unpack(object.__originalBorderRGB))
				end)
				
				object._hasHoverScripts = true
			end
		end

		local SetShineScripts = function(object, shineAlpha, shineDuration, shineScale)
			if not(object.__hasShineEffect) then
				object.shineFrame = F.Shine:New(object, shineAlpha, shineDuration, shineScale)
				
				object:HookScript("OnEnter", function(self) 
					self.shineFrame:Start()
				end)
			
				object.__hasShineEffect = true
			end
		end
		local backdropFrames = {}
		local glowFrames = {}
		local templates = {
			tinyglow = function(object, ...)
				if (glowFrames[object]) then return glowFrames[object] end
				local r, g, b, a = ...
				local glow = makeBackdropFrame(object, 3)
				glow:SetBackdrop(M("Backdrop", "TinyGlow"))
				glow:SetBackdropBorderColor(r or 0, g or 0, b or 0, a or 0.75)
				glowFrames[object] = glow
				return glowFrames[object]
			end;

			smallglow = function(object, ...)
				if (glowFrames[object]) then return glowFrames[object] end
				local r, g, b, a = ...
				local glow = makeBackdropFrame(object, 5)
				glow:SetBackdrop(M("Backdrop", "SmallGlow"))
				glow:SetBackdropBorderColor(r or 0, g or 0, b or 0, a or 0.75)
				glowFrames[object] = glow
				return glowFrames[object]
			end;
			
			largeglow = function(object, ...)
				if (glowFrames[object]) then return glowFrames[object] end
				local r, g, b, a = ...
				local glow = makeBackdropFrame(object, 16)
				glow:SetBackdrop(M("Backdrop", "Glow"))
				glow:SetBackdropBorderColor(r or 0, g or 0, b or 0, a or 0.35)
				glowFrames[object] = glow
				return glowFrames[object]
			end;
			
			hugeglow = function(object, ...)
				if (glowFrames[object]) then return glowFrames[object] end
				local r, g, b, a = ...
				local glow = makeBackdropFrame(object, 32)
				glow:SetBackdrop(M("Backdrop", "HugeGlow"))
				glow:SetBackdropBorderColor(r or 0, g or 0, b or 0, a or 0.35)
				glowFrames[object] = glow
				return glowFrames[object]
			end;
			
			smallindented = function(object, ...)
				if (glowFrames[object]) then return glowFrames[object] end
				local r, g, b, a = ...
				local glow = makeBackdropFrame(object, 5-2) -- the indent is 2px
				glow:SetBackdrop(M("Backdrop", "SmallGlow"))
				glow:SetBackdropBorderColor(r or 0, g or 0, b or 0, a or 0.55)
				glowFrames[object] = glow
				return glowFrames[object]
			end;

			backdrop = function(object, ...)
				object:SetBackdrop(M("Backdrop", "PixelBorder-Blank"))
				object:SetBackdropColor(C.background[1], C.background[2], C.background[3], gUI:GetPanelAlpha())
				object:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				backdropFrames[object] = object
				return backdropFrames[object]
			end;
			
			outerbackdrop = function(object, ...)
				if (backdropFrames[object]) then return backdropFrames[object] end
				local border = makeBackdropFrame(object, 3, 1) 
				border:SetBackdrop(M("Backdrop", "PixelBorder-Satin")) -- just a border, no background
				border:SetBackdropColor(C.background[1], C.background[2], C.background[3], gUI:GetPanelAlpha())
				border:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				
				-- emulating the v2 functionality for skinning here
				local target, top, left, bottom, right = ...
				border:SetPoint("TOPLEFT", target or object, "TOPLEFT", -3 + (left or 0), 3 - (top or 0))
				border:SetPoint("BOTTOMRIGHT", target or object, "BOTTOMRIGHT", 3 - (right or 0), -3 + (bottom or 0))
				
				backdropFrames[object] = border -- index by target object, not the borderframe

				return border
			end;
			
			insetbackdrop = function(object, ...)
				object:SetBackdrop(M("Backdrop", "PixelBorder-Blank-Indented"))
				object:SetBackdropColor(C.background[1], C.background[2], C.background[3], gUI:GetPanelAlpha())
				object:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				backdropFrames[object] = object
				return backdropFrames[object]
			end;
			
			satinbackdrop = function(object, ...)
				object:SetBackdrop(M("Backdrop", "PixelBorder-Satin"))
				object:SetBackdropColor(C.background[1], C.background[2], C.background[3], gUI:GetPanelAlpha())
				object:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				backdropFrames[object] = object
				return backdropFrames[object]
			end;
			
			insetsatinbackdrop = function(object, ...)
				object:SetBackdrop(M("Backdrop", "PixelBorder-Satin-Indented"))
				object:SetBackdropColor(C.background[1], C.background[2], C.background[3], gUI:GetPanelAlpha())
				object:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				backdropFrames[object] = object
				return backdropFrames[object]
			end;
			-- 
			targetborder = function(object, ...)
				if (backdropFrames[object]) then return backdropFrames[object] end
				local border = makeBackdropFrame(object, 0, 1) 
				border:SetBackdrop(M("Backdrop", "TargetBorder")) -- just a border, no background
				-- border:SetBackdropColor(C.background[1], C.background[2], C.background[3], gUI:GetPanelAlpha())
				border:SetBackdropBorderColor(C.target[1], C.target[2], C.target[3], 1)
				backdropFrames[object] = border -- index by target object, not the borderframe
				return border
			end;
			
			gloss = function(object, ...)
				local target = ...
				object.Gloss = object.Gloss or object:CreateTexture()
				object.Gloss:SetDrawLayer("OVERLAY", 2)
				object.Gloss:SetTexture(M("Background", "gUI™ Gloss"))
				object.Gloss:SetVertexColor(1, 1, 1, 1/4)
				object.Gloss:SetAllPoints(target or object)
				return object.Gloss
			end;
			
			shade = function(object, ...)
				local target = ...
				object.Shade = object.Shade or object:CreateTexture()
				object.Shade:SetDrawLayer("OVERLAY", 1)
				object.Shade:SetTexture(M("Background", "gUI™ Shade"))
				object.Shade:SetVertexColor(0, 0, 0, 1/4)
				object.Shade:SetAllPoints(target or object)
				return object.Shade
			end;
			
			simpleborder = function(object, ...)
				object:SetBackdrop(M("Backdrop", "PixelBorder"))
				object:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				backdropFrames[object] = object
				return backdropFrames[object]
			end;
			
			border = function(object, ...)
				if (backdropFrames[object]) then return backdropFrames[object] end
				local border = makeBackdropFrame(object, 3, -1) -- put this above the object
				border:SetBackdrop(M("Backdrop", "PixelBorder")) -- just a border, no background
				border:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				backdropFrames[object] = border -- index by target object, not the borderframe
				return border
			end;
			
			-- creates and returns a castbar backdrop with a border, and puts it around the object
			castbarwithborder = function(object, ...)
				if (backdropFrames[object]) then return backdropFrames[object] end
				local border = makeBackdropFrame(object, 3, 2) 
				border:SetBackdrop(M("Backdrop", "StatusBarBorder")) 
				border:SetBackdropColor(C.overlay[1], C.overlay[2], C.overlay[3], 1)
				border:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				backdropFrames[object] = border -- index by target object, not the borderframe
				return border
			end;

			-- sets the backdrop of object to be a castbar with a border. content must be 3px indented
			castbarbackdropwithborder = function(object, ...)
				if (backdropFrames[object]) then return backdropFrames[object] end
				-- local border = makeBackdropFrame(object, 3, -1) -- put this above the object
				object:SetBackdrop(M("Backdrop", "StatusBarBorder")) 
				object:SetBackdropColor(C.overlay[1], C.overlay[2], C.overlay[3], 1)
				object:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				backdropFrames[object] = object -- index by target object, not the borderframe
				return backdropFrames[object]
			end;

			itembackdrop = function(object, ...)
				if (backdropFrames[object]) then return backdropFrames[object] end
				object:SetFrameLevel(object:GetFrameLevel() + 2)
				local border = makeBackdropFrame(object, 3, 2) 
				border:SetBackdrop(M("Backdrop", "PixelBorder-Satin")) 
				border:SetBackdropColor(C.backdrop[1], C.backdrop[2], C.backdrop[3], gUI:GetPanelAlpha())
				border:SetBackdropBorderColor(C.border[1], C.border[2], C.border[3], 1)
				
				-- emulating the v2 functionality for skinning here
				local target, top, left, bottom, right = ...
				border:SetPoint("TOPLEFT", target or object, "TOPLEFT", -3 + (left or 0), 3 - (top or 0))
				border:SetPoint("BOTTOMRIGHT", target or object, "BOTTOMRIGHT", 3 - (right or 0), -3 + (bottom or 0))
				gUI:CreateUIShadow(border)
				
				backdropFrames[object] = border -- index by target object, not the borderframe
				return backdropFrames[object]
			end;
			
			--------------------------------------------------------------------------------------------------
			--		The old element skinner templates
			--------------------------------------------------------------------------------------------------
			button = function(object, strip, shine, shineAlpha, shineDuration, shineScale)
				if (strip) then 
					gUI:DisableTextures(object) 
				end

				gUI:SetUITemplate(object, "backdrop")

				if not(object.candyLayer) then
					object.candyLayer = CreateFrame("Frame", nil, object)
					object.candyLayer:SetPoint("TOPLEFT", 3, -3)
					object.candyLayer:SetPoint("BOTTOMRIGHT", -3, 3)
					gUI:SetUITemplate(object.candyLayer, "gloss")
					gUI:SetUITemplate(object.candyLayer, "shade")
				end
				
				local text = object.text or object:GetName() and _G[object:GetName() .. "Text"]
				-- if (text) and (text.SetFontObject) then
					-- text:SetFontObject(gUI_TextFontSmall)
					-- object:SetNormalFontObject(gUI_TextFontSmall)
					-- object:SetHighlightFontObject(gUI_TextFontSmallWhite)
					-- object:SetDisabledFontObject(gUI_TextFontSmallDisabled)
				-- end
				
				if (object:GetName()) then
					local l = _G[object:GetName() .. "Left"]
					local m = _G[object:GetName() .. "Middle"]
					local r = _G[object:GetName() .. "Right"]
					
					if (l) then l:SetAlpha(0) end
					if (m) then m:SetAlpha(0) end
					if (r) then r:SetAlpha(0) end
				end

				if (object.SetNormalTexture) then object:SetNormalTexture("") end
				if (object.SetHighlightTexture) then object:SetHighlightTexture("") end
				if (object.SetPushedTexture) then object:SetPushedTexture("") end
				if (object.SetDisabledTexture) then object:SetDisabledTexture("") end
				
				SetHoverScripts(object)
				
				if (shine) then
					SetShineScripts(object, shineAlpha, shineDuration, shineScale)
				end
				
				return object
			end;

			["button-indented"] = function(object, strip, shine, shineAlpha, shineDuration, shineScale)
				if (strip) then 
					gUI:DisableTextures(object) 
				end

				gUI:SetUITemplate(object, "insetbackdrop")

				object.candyLayer = CreateFrame("Frame", nil, object)
				object.candyLayer:SetPoint("TOPLEFT", 5, -5)
				object.candyLayer:SetPoint("BOTTOMRIGHT", -5, 5)
				gUI:SetUITemplate(object.candyLayer, "gloss")
				gUI:SetUITemplate(object.candyLayer, "shade")
				
				local text = object.text or object:GetName() and _G[object:GetName() .. "Text"]
				-- if (text) and (text.SetFontObject) then
					-- text:SetFontObject(gUI_TextFontSmall)
					-- object:SetNormalFontObject(gUI_TextFontSmall)
					-- object:SetHighlightFontObject(gUI_TextFontSmallWhite)
					-- object:SetDisabledFontObject(gUI_TextFontSmallDisabled)
				-- end

				if (object:GetName()) then
					local l = _G[object:GetName() .. "Left"]
					local m = _G[object:GetName() .. "Middle"]
					local r = _G[object:GetName() .. "Right"]
					
					if (l) then l:SetAlpha(0) end
					if (m) then m:SetAlpha(0) end
					if (r) then r:SetAlpha(0) end
				end

				if (object.SetNormalTexture) then object:SetNormalTexture("") end
				if (object.SetHighlightTexture) then object:SetHighlightTexture("") end
				if (object.SetPushedTexture) then object:SetPushedTexture("") end
				if (object.SetDisabledTexture) then object:SetDisabledTexture("") end
				
				SetHoverScripts(object)

				if (shine) then
					SetShineScripts(object, shineAlpha, shineDuration, shineScale)
				end
				
				return object
			end;

			tab = function(object, strip, shine, shineAlpha, shineDuration, shineScale)
				if (strip) then 
					gUI:DisableTextures(object) 
				end

				local backdrop = gUI:SetUITemplate(object, "outerbackdrop", nil, 5, 12, 5, 12)

				if not(object.candyLayer) then
					object.candyLayer = CreateFrame("Frame", nil, backdrop)
					object.candyLayer:SetPoint("TOPLEFT", 3, -3)
					object.candyLayer:SetPoint("BOTTOMRIGHT", -3, 3)
					gUI:SetUITemplate(object.candyLayer, "gloss")
					gUI:SetUITemplate(object.candyLayer, "shade")
				end
				
				local text = object.text or object:GetName() and _G[object:GetName() .. "Text"]
				-- if (text) and (text.SetFontObject) then
					-- text:SetFontObject(gUI_TextFontSmall)
					-- object:SetNormalFontObject(gUI_TextFontSmall)
					-- object:SetHighlightFontObject(gUI_TextFontSmallWhite)
					-- object:SetDisabledFontObject(gUI_TextFontSmallDisabled)
					
					-- not strictly happy with this, but it avoids taint, if nothing else
					--[[
					hooksecurefunc("PanelTemplates_SelectTab = function(tab) 
						if (tab == object) then
							tab:SetDisabledFontObject(GUIS_SystemFontSmallHighlight)
						end
					end)
					
					hooksecurefunc("PanelTemplates_SetDisabledTabState = function(tab) 
						if (tab == object) then
							tab:SetDisabledFontObject(GUIS_SystemFontSmallDisabled)
						end
					end)
					]]--
				-- end

				if (object:GetName()) then
					local l = _G[object:GetName() .. "Left"]
					local m = _G[object:GetName() .. "Middle"]
					local r = _G[object:GetName() .. "Right"]
					local ld = _G[object:GetName() .. "LeftDisabled"]
					local md = _G[object:GetName() .. "MiddleDisabled"]
					local rd = _G[object:GetName() .. "RightDisabled"]
					
					if (l) then l:SetAlpha(0) end
					if (m) then m:SetAlpha(0) end
					if (r) then r:SetAlpha(0) end
					if (ld) then ld:SetAlpha(0) end
					if (md) then md:SetAlpha(0) end
					if (rd) then rd:SetAlpha(0) end
				end

				if (object.SetNormalTexture) then object:SetNormalTexture("") end
				if (object.SetHighlightTexture) then object:SetHighlightTexture("") end
				if (object.SetPushedTexture) then object:SetPushedTexture("") end
				if (object.SetDisabledTexture) then object:SetDisabledTexture("") end
				
				SetHoverScripts(object)

				if (shine) then
					SetShineScripts(object, shineAlpha, shineDuration, shineScale)
				end
				
				return object
			end;

			scrollbar = function(object, strip)
				-- gUI:DisableTextures(object)
				gUI:RemoveClutter(object)
				
				local name = object:GetName()
				if (name) then
					if (_G[name .. "ScrollDownButton"]) then
						gUI:SetUITemplate(_G[name .. "ScrollDownButton"], "arrow", "down")
						_G[name .. "ScrollDownButton"]:GetNormalTexture():SetTexCoord(0, 1, 0, 1)
						_G[name .. "ScrollDownButton"]:GetPushedTexture():SetTexCoord(0, 1, 0, 1)
						_G[name .. "ScrollDownButton"]:GetDisabledTexture():SetTexCoord(0, 1, 0, 1)
						_G[name .. "ScrollDownButton"]:GetHighlightTexture():SetTexCoord(0, 1, 0, 1)
					end

					if (_G[name .. "ScrollUpButton"]) then
						gUI:SetUITemplate(_G[name .. "ScrollUpButton"], "arrow", "up")
						_G[name .. "ScrollUpButton"]:GetNormalTexture():SetTexCoord(0, 1, 0, 1)
						_G[name .. "ScrollUpButton"]:GetPushedTexture():SetTexCoord(0, 1, 0, 1)
						_G[name .. "ScrollUpButton"]:GetDisabledTexture():SetTexCoord(0, 1, 0, 1)
						_G[name .. "ScrollUpButton"]:GetHighlightTexture():SetTexCoord(0, 1, 0, 1)
					end
					
					if (_G[name .. "ThumbTexture"]) then
						_G[name .. "ThumbTexture"]:SetTexture(M("Background", "Blank"), 1) 
						_G[name .. "ThumbTexture"]:SetVertexColor(C["disabled"][1], C["disabled"][2], C["disabled"][3], 1) 
						_G[name .. "ThumbTexture"]:SetWidth(_G[name .. "ThumbTexture"]:GetWidth() - 8)
						-- _G[name .. "ThumbTexture"]:SetHeight(_G[name .. "ThumbTexture"]:GetHeight() - 8)
						
						gUI:CreateUIShadow(object)
						gUI:GetUIShadow(object):ClearAllPoints()
						gUI:GetUIShadow(object):SetPoint("TOPLEFT", _G[name .. "ThumbTexture"], "TOPLEFT", -3, 3)
						gUI:GetUIShadow(object):SetPoint("BOTTOMRIGHT", _G[name .. "ThumbTexture"], "BOTTOMRIGHT", 3, -3)
					end
				end
				
				for i = 1, object:GetNumRegions() do
					local region = select(i, object:GetRegions())
					
					if (region:GetName()) and ((region:GetName():find("Up")) or (region:GetName():find("Down"))) then
						gUI:SetUITemplate(region, "arrow")
						region:GetNormalTexture():SetTexCoord(0, 1, 0, 1)
						region:GetPushedTexture():SetTexCoord(0, 1, 0, 1)
						region:GetDisabledTexture():SetTexCoord(0, 1, 0, 1)
						region:GetHighlightTexture():SetTexCoord(0, 1, 0, 1)
					end
				end
				
				for v = 1, object:GetNumChildren() do
					local child = select(v, object:GetChildren())
					
					for i = 1, child:GetNumRegions() do
						local region = select(i, child:GetRegions())
						
						if (region:GetName()) and ((region:GetName():find("Up")) or (region:GetName():find("Down"))) then
							gUI:SetUITemplate(region, "arrow")
							region:GetNormalTexture():SetTexCoord(0, 1, 0, 1)
							region:GetPushedTexture():SetTexCoord(0, 1, 0, 1)
							region:GetDisabledTexture():SetTexCoord(0, 1, 0, 1)
							region:GetHighlightTexture():SetTexCoord(0, 1, 0, 1)
						end
					end
				end
				
				return object
			end;

			slider = function(object, orientation)
				gUI:SetUITemplate(object, "insetbackdrop")
				object.SetBackdrop = noop
				return object
			end;

			closebutton = function(object, ...)
				object:SetNormalTexture(M("Button", "gUI™ CloseButton"))
				object:SetPushedTexture(M("Button", "gUI™ CloseButtonDown"))
				object:SetHighlightTexture(M("Button", "gUI™ CloseButtonHighlight"))
				object:SetDisabledTexture(M("Button", "gUI™ CloseButtonDisabled"))

				object:SetSize(16, 16)

				if (...) then
					local a, b, c, d, e, f = ...
					
					if (a == true) then
						object.SetNormalTexture = noop
						object.SetPushedTexture = noop
						object.SetHighlightTexture = noop
						object.SetDisabledTexture = noop
						
						if (b) then
							object:ClearAllPoints()
							object:SetPoint(b, c, d, e, f)
						end
					else
						object:ClearAllPoints()
						object:SetPoint(a, b, c, d, e)
					end
				else
					local point, anchor, relpoint, x, y = object:GetPoint()
					if (point) then
						object:ClearAllPoints()
						object:SetPoint(point, anchor, relpoint, x-8, y-8)
					end
				end
				
				return object
			end;

			radiobutton = function(object, strip)
				if (strip) then 
					gUI:DisableTextures(object) 
				end

				gUI:SetUITemplate(object, "backdrop"):SetBackdropColor(nil, nil, nil, gUI:GetOverlayAlpha())
				
				local name = object:GetName()
				local text = object.text or name and _G[name .. "Text"]
				-- if (text) and (text.SetFontObject) then
					-- text:SetFontObject(gUI_TextFontSmall)
					-- object:SetNormalFontObject(gUI_TextFontSmall)
					-- object:SetHighlightFontObject(gUI_TextFontSmallWhite)
					-- object:SetDisabledFontObject(gUI_TextFontSmallDisabled)
				-- end

				if (object.SetCheckedTexture) then
					object:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check")
					
					object:GetCheckedTexture():SetDrawLayer("OVERLAY")
					object:GetCheckedTexture():SetTexCoord(0, 1, 0, 1)
					object:GetCheckedTexture().SetTexCoord = noop
				end
				
				if (object.SetDisabledTexture) then
					object:SetDisabledTexture("Interface\\Buttons\\UI-CheckBox-Check-Disabled")

					object:GetDisabledTexture():SetDrawLayer("OVERLAY")
					object:GetDisabledTexture():SetTexCoord(0, 1, 0, 1)
					object:GetDisabledTexture().SetTexCoord = noop
				end
				
				if (object.SetHighlightTexture) then
					local highlight = object:CreateTexture(nil, "OVERLAY")
					highlight:ClearAllPoints()
					highlight:SetPoint("TOPLEFT", object, "TOPLEFT", 2, -2)
					highlight:SetPoint("BOTTOMRIGHT", object, "BOTTOMRIGHT", -2, 2)
					highlight:SetTexture(C["index"][1], C["index"][2], C["index"][3], 1/4)

					object:SetHighlightTexture(highlight)

					if (name) then
						_G[name .. "Highlight"] = highlight
					end
				end
				
				object.SetDisabledTexture = noop	
				object.SetNormalTexture = noop
				object.SetPushedTexture = noop
				object.SetHighlightTexture = noop
				
				return object
			end;

			-- jumping through hoops to make this nonsense compatible with ScrollFrames
			dropdown = function(object, strip, width)
				if (strip) then
					if (strip == true) then
						gUI:DisableTextures(object)
					end
					
					if (tonumber(strip)) and (tonumber(strip) > 0) then
						width = strip
					end
				end
				
				local button = _G[object:GetName() .. "Button"]
				local text = _G[object:GetName() .. "Text"]

				if (width) then
					object:SetWidth(width)
				end
				
				gUI:SetUITemplate(object, "insetbackdrop", button):SetBackdropColor(nil, nil, nil, gUI:GetOverlayAlpha())
				gUI:SetUITemplate(button, "arrow", "down")
				
				text:ClearAllPoints()
				text:SetPoint("RIGHT", button:GetNormalTexture(), "LEFT", 0, 0)
				text:SetPoint("LEFT", object, "LEFT", 8, 0)
				-- text:SetPoint("BOTTOM", button, "BOTTOM", 0, 5)
				text:SetJustifyV("MIDDLE")
				text.ClearAllPoints = noop
				text.SetPoint = noop
				
				text:SetParent(button)
				text:SetDrawLayer("OVERLAY")
				-- text:SetFontObject(gUI_TextFontSmallWhite)
				
				button:ClearAllPoints()
				button:SetPoint("RIGHT", object, "RIGHT", -8, 0)
				button:SetPoint("LEFT", object, "LEFT", 8, 0)
				button:SetPoint("TOP", object, "TOP", 0, 0)
				button:SetPoint("BOTTOM", object, "BOTTOM", 0, 0)
				button.ClearAllPoints = noop
				button.SetPoint = noop
				
				return button
			end;

			checkbutton = function(object, ...)
				gUI:DisableTextures(object)
				gUI:SetUITemplate(object, "insetbackdrop")
				
				local text = object.text or object:GetName() and _G[object:GetName() .. "Text"]
				-- if (text) and (text.SetFontObject) then
					-- text:SetFontObject(gUI_TextFontSmall)
					-- object:SetNormalFontObject(gUI_TextFontSmall)
					-- object:SetHighlightFontObject(gUI_TextFontSmallWhite)
					-- object:SetDisabledFontObject(gUI_TextFontSmallDisabled)
				-- end

				if (object.SetCheckedTexture) then
					object:SetCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check")
					object:GetCheckedTexture():SetTexCoord(0, 1, 0, 1)
					object:GetCheckedTexture().SetTexCoord = noop
					object:GetCheckedTexture():SetPoint("TOPLEFT", -3, 3)
					object:GetCheckedTexture():SetPoint("BOTTOMRIGHT", 3, -3)
					object:GetCheckedTexture():SetParent(object)
					object:GetCheckedTexture():SetDrawLayer("OVERLAY", 7)
				end
				
				if (object.SetDisabledTexture) then
					object:SetDisabledTexture("")
					object:GetDisabledTexture():SetTexCoord(0, 1, 0, 1)
					object:GetDisabledTexture().SetTexCoord = noop
				end

				if (object.SetDisabledCheckedTexture) then
					object:SetDisabledCheckedTexture("Interface\\Buttons\\UI-CheckBox-Check-Disabled")
					object:GetDisabledCheckedTexture():SetTexCoord(0, 1, 0, 1)
					object:GetDisabledCheckedTexture().SetTexCoord = noop
				end
				
				if (object.SetHighlightTexture) then
					local highlight = object:CreateTexture()
					highlight:SetDrawLayer("OVERLAY")
					highlight:ClearAllPoints()
					highlight:SetPoint("TOPLEFT", object, "TOPLEFT", 5, -5)
					highlight:SetPoint("BOTTOMRIGHT", object, "BOTTOMRIGHT", -5, 5)
					highlight:SetTexture(C["index"][1], C["index"][2], C["index"][3], 1/4)

					object:SetHighlightTexture(highlight)

					if (object:GetName()) then
						_G[object:GetName() .. "Highlight"] = highlight
					end
				end
				
				object.SetNormalTexture = noop
				object.SetPushedTexture = noop
				object.SetHighlightTexture = noop
				
				return object
			end;

			editbox = function(object, top, left, bottom, right, textTop, textLeft, textBottom, textRight)
				gUI:RemoveClutter(object)
				
				object:SetFrameLevel(object:GetFrameLevel() + 1)

				local backdrop = gUI:SetUITemplate(object, "outerbackdrop")
				backdrop:ClearAllPoints()
				backdrop:SetPoint("TOPLEFT", object, "TOPLEFT", -3 + (left or 0), 2 + (top or 0))
				backdrop:SetPoint("BOTTOMRIGHT", object, "BOTTOMRIGHT", 3+ (right or 0), -2 + (bottom or 0))

				-- if (object.SetFontObject) then
					-- object:SetFontObject(gUI_TextFontNormal)
				-- end

				local name = object:GetName()
				if (name) then
					if ((name:find("Silver")) or (name:find("Copper"))) then
						backdrop:SetPoint("BOTTOMRIGHT", object, "BOTTOMRIGHT", -12, -2 + (bottom or 0))
					end
					
					-- local header = _G[name.."Header"]
					-- if (header) then
						-- header:SetFontObject(gUI_TextFontNormal)
					-- end
				end
				
				if (textTop) or (textLeft) or (textBottom) or (textRight) then
					object:SetTextInsets(textLeft or 0, textRight or 0, textTop or 0, textBottom or 0)
				end
				
				return backdrop
			end;

			statusbar = function(object, border)
				gUI:DisableTextures(object)
				
				if (border) then
					local backdrop = gUI:SetUITemplate(object, "border")
				end

				object:SetStatusBarTexture(M("Statusbar", "gUI™ StatusBar"))
				gUI:SetUITemplate(object, "gloss")
				gUI:SetUITemplate(object, "shade")
				
				-- local name = object:GetName()

				-- local text = object.text or name and _G[name .. "Text"]
				-- if (text) and (text.SetFontObject) then
					-- text:SetFontObject(gUI_DisplayFontSmallOutlineWhite)
				-- end

				-- local title = object.title or name and _G[name .. "Title"]
				-- if (title) and (title.SetFontObject) then
					-- title:SetFontObject(gUI_DisplayFontSmallOutlineWhite)
				-- end

				-- local label = object.label or name and _G[name .. "Label"]
				-- if (label) and (label.SetFontObject) then
					-- label:SetFontObject(gUI_DisplayFontSmallOutlineWhite)
				-- end
				
				return object
			end;

			-- possibly the biggest overkill ever, just to save me time later
			-- lazy is the new art, at least when I do it
			arrow = function(object, direction, ...)
				local Direction

				if not(direction) then
					local name = (object.GetName) and (object:GetName())
					if (name:find("Collapse")) or (name:find("Toggle")) then
						if not(object.BlizzSetNormalTexture) then
							object.BlizzSetNormalTexture = object.SetNormalTexture
							object.BlizzSetPushedTexture = object.SetPushedTexture
							object.BlizzSetHighlightTexture = object.SetHighlightTexture
							object.BlizzSetDisabledTexture = object.SetDisabledTexture
						end
						
						object.SetNormalTexture = function(object, normal)
							if not(normal) or not(normal.find) then
								return
							end
							
							if (normal and normal:find("Plus")) then
								object:BlizzSetNormalTexture(M("Button", "gUI™ ArrowDown"))
								object:BlizzSetPushedTexture(M("Button", "gUI™ ArrowDown"))
								object:BlizzSetHighlightTexture(M("Button", "gUI™ ArrowDownHighlight"))
								object:BlizzSetDisabledTexture(M("Button", "gUI™ ArrowDownDisabled"))
														
							elseif (normal and normal:find("Minus")) then
								object:BlizzSetNormalTexture(M("Button", "gUI™ ArrowUp"))
								object:BlizzSetPushedTexture(M("Button", "gUI™ ArrowUp"))
								object:BlizzSetHighlightTexture(M("Button", "gUI™ ArrowUpHighlight"))
								object:BlizzSetDisabledTexture(M("Button", "gUI™ ArrowUpDisabled"))
							end
						end
						
						local normal = object:GetNormalTexture():GetTexture()
						
						object.SetPushedTexture = noop
						object.SetHighlightTexture = noop
						object.SetDisabledTexture = noop
						object:SetNormalTexture(normal)
						
						Direction = (normal and normal:find("Plus")) and "Down" or "Up"
						
					elseif (name:find("Expand")) then
						if not(object.BlizzSetNormalTexture) then
							object.BlizzSetNormalTexture = object.SetNormalTexture
							object.BlizzSetPushedTexture = object.SetPushedTexture
							object.BlizzSetHighlightTexture = object.SetHighlightTexture
							object.BlizzSetDisabledTexture = object.SetDisabledTexture
						end

						object.SetNormalTexture = function(object, normal) 
							if not(normal) or not(normal.find) then
								return
							end
							
							if (normal and normal:find("Prev")) then
								object:BlizzSetNormalTexture(M("Button", "gUI™ ArrowLeft"))
								object:BlizzSetPushedTexture(M("Button", "gUI™ ArrowLeft"))
								object:BlizzSetHighlightTexture(M("Button", "gUI™ ArrowLeftHighlight"))
								object:BlizzSetDisabledTexture(M("Button", "gUI™ ArrowLeftDisabled"))
														
							elseif (normal and normal:find("Next")) then
								object:BlizzSetNormalTexture(M("Button", "gUI™ ArrowRight"))
								object:BlizzSetPushedTexture(M("Button", "gUI™ ArrowRight"))
								object:BlizzSetHighlightTexture(M("Button", "gUI™ ArrowRightHighlight"))
								object:BlizzSetDisabledTexture(M("Button", "gUI™ ArrowRightDisabled"))
							end
						end

						local normal = object:GetNormalTexture():GetTexture()
						
						object.SetPushedTexture = noop
						object.SetHighlightTexture = noop
						object.SetDisabledTexture = noop
						object:SetNormalTexture(normal)
						Direction = (normal and normal:find("Prev")) and "Left" or "Right"
						
					elseif (name:find("Prev")) or (name:find("Left")) or (name:find("Dec")) then Direction = "Left"
					elseif (name:find("Next")) or (name:find("Right")) or (name:find("Inc")) then Direction = "Right"
					elseif (name:find("Up")) then Direction = "Up"
					elseif (name:find("Down")) then Direction = "Down"
					elseif (name:find("Top")) then Direction = "Top"
					elseif (name:find("Bottom")) then Direction = "Bottom"
					elseif (name:find("First")) then Direction = "First"
					elseif (name:find("Last")) then Direction = "Last" 
					end
				else
					if (direction == "left") then Direction = "Left"
					elseif (direction == "right") then Direction = "Right"
					elseif (direction == "up") then Direction = "Up"
					elseif (direction == "down") then Direction = "Down"
					elseif (direction == "top") then Direction = "Top"
					elseif (direction == "bottom") then Direction = "Bottom"
					elseif (direction == "first") then Direction = "First"
					elseif (direction == "collapse") then
						if not(object.BlizzSetNormalTexture) then
							object.BlizzSetNormalTexture = object.SetNormalTexture
							object.BlizzSetPushedTexture = object.SetPushedTexture
							object.BlizzSetHighlightTexture = object.SetHighlightTexture
							object.BlizzSetDisabledTexture = object.SetDisabledTexture
						end
						
						object.SetNormalTexture = function(object, normal)
							if not(normal) or not(normal.find) then
								return
							end
							
							if (normal and normal:find("Plus")) then
								object:BlizzSetNormalTexture(M("Button", "gUI™ ArrowDown"))
								object:BlizzSetPushedTexture(M("Button", "gUI™ ArrowDown"))
								object:BlizzSetHighlightTexture(M("Button", "gUI™ ArrowDownHighlight"))
								object:BlizzSetDisabledTexture(M("Button", "gUI™ ArrowDownDisabled"))
														
							elseif (normal and normal:find("Minus")) then
								object:BlizzSetNormalTexture(M("Button", "gUI™ ArrowUp"))
								object:BlizzSetPushedTexture(M("Button", "gUI™ ArrowUp"))
								object:BlizzSetHighlightTexture(M("Button", "gUI™ ArrowUpHighlight"))
								object:BlizzSetDisabledTexture(M("Button", "gUI™ ArrowUpDisabled"))
							end
						end
						
						local normal = object:GetNormalTexture():GetTexture()
						
						object.SetPushedTexture = noop
						object.SetHighlightTexture = noop
						object.SetDisabledTexture = noop
						object:SetNormalTexture(normal)
						
						Direction = (normal and normal:find("Plus")) and "Down" or "Up"
						
					elseif (direction == "expand") then
						if not(object.BlizzSetNormalTexture) then
							object.BlizzSetNormalTexture = object.SetNormalTexture
							object.BlizzSetPushedTexture = object.SetPushedTexture
							object.BlizzSetHighlightTexture = object.SetHighlightTexture
							object.BlizzSetDisabledTexture = object.SetDisabledTexture
						end

						object.SetNormalTexture = function(object, normal) 
							if not(normal) or not(normal.find) then
								return
							end
							
							if (normal and normal:find("Prev")) then
								object:BlizzSetNormalTexture(M("Button", "gUI™ ArrowLeft"))
								object:BlizzSetPushedTexture(M("Button", "gUI™ ArrowLeft"))
								object:BlizzSetHighlightTexture(M("Button", "gUI™ ArrowLeftHighlight"))
								object:BlizzSetDisabledTexture(M("Button", "gUI™ ArrowLeftDisabled"))
														
							elseif (normal and normal:find("Next")) then
								object:BlizzSetNormalTexture(M("Button", "gUI™ ArrowRight"))
								object:BlizzSetPushedTexture(M("Button", "gUI™ ArrowRight"))
								object:BlizzSetHighlightTexture(M("Button", "gUI™ ArrowRightHighlight"))
								object:BlizzSetDisabledTexture(M("Button", "gUI™ ArrowRightDisabled"))
							end
						end

						local normal = object:GetNormalTexture():GetTexture()
						
						object.SetPushedTexture = noop
						object.SetHighlightTexture = noop
						object.SetDisabledTexture = noop
						object:SetNormalTexture(normal)
						
						Direction = (normal and normal:find("Prev")) and "Left" or "Right"
					
					elseif (direction == "last") then
						Direction = "Last"
					end
				end

				if not(Direction) then
					return
				end
				
				gUI:DisableTextures(object)

				object:SetNormalTexture(M("Button", "gUI™ Arrow" .. Direction))
				object:SetPushedTexture(M("Button", "gUI™ Arrow" .. Direction))
				object:SetHighlightTexture(M("Button", "gUI™ Arrow" .. Direction .. "Highlight"))
				object:SetDisabledTexture(M("Button", "gUI™ Arrow" .. Direction .. "Disabled"))
				
				if (...) then
					object:ClearAllPoints()
					object:SetPoint(...)
				end
				
				return object
			end;
		}
		
		-- same rules as in v2 apply;
		-- 	*the function must not create duplicate items
		-- 	*the function must return an object which responds to backdrop commands
		gUI.RegisterUITemplate = function(self, name, func)
			if not(templates[name]) then
				templates[name] = func
			end
		end
		
		-- debugging in my blizzard skin conversions require this
		local prev
		gUI.SetUITemplate = function(self, object, template, ...)
			if not(object) then
				print("SetUITemplate: The object after " .. (prev or "nil") .. " with the template " .. template .. "doesn't exist!")
				return
			else
				prevName = object:GetName()
			end

			self:argCheck(object, 1, "table")
			self:argCheck(template, 2, "string")
			
			if (templates[template]) then
				return templates[template](object, ...)
			else
				print(("The requested template '%s' does not exist!"):format(template))
			end
		end
		
		-- this seems a bit backwards. but oh well.
		gUI.GetBackgroundColor = function(self)
			return unpack(C.backdrop)
		end

		gUI.GetBackdropColor = function(self)
			return unpack(C.background)
		end

		gUI.GetBackdropBorderColor = function(self)
			return unpack(C.border)
		end

		gUI.GetPanelAlpha = function(self)
			return defaults.panelAlpha -- don't currently want this meddled with
		end
		
		gUI.GetOverlayAlpha = function(self)
			return db.overlayAlpha
		end

		gUI.SetShadeAlpha = function(self, a)
			db.shadeAlpha = a
		end

		gUI.GetShadeAlpha = function(self)
			return db.shadeAlpha
		end

		gUI.SetGlossAlpha = function(self, a)
			db.glossAlpha = a
		end

		gUI.GetGlossAlpha = function(self, a)
			return db.glossAlpha
		end

		-- the following aren't really configurable, 
		-- but fit in here anyway as they represent global styling for the UI
		gUI.GetGlossTexture = function(self)
			return M("Background", "gUI™ Gloss")
		end

		gUI.GetShadeTexture = function(self)
			return M("Background", "gUI™ Shade")
		end

		gUI.GetItemButtonBackdrop = function(self)
			return M("Backdrop", "ItemButton")
		end

		gUI.GetBlankTexture = function(self)
			return M("Background", "Blank")
		end

		-- setup function to initialize the buttonstyling
		-- any module that needs buttonstyling should call this, 
		-- as there is no guarantee that any other have
		gUI.SetUpButtonStyling = function(self)
			gABT:SetBackdrop(gUI:GetItemButtonBackdrop()) -- set the default backdrop for all buttons
			gABT:SetBorderSize(3) -- indent for icons, textures and button content
			gABT:SetTexture(gUI:GetBlankTexture()) -- unneeded, it uses it already
			gABT:SetGlossTexture(gUI:GetGlossTexture())
			gABT:SetShadeTexture(gUI:GetShadeTexture())
			gABT:SetGlossAlpha(gUI:GetGlossAlpha())
			gABT:SetShadeAlpha(gUI:GetShadeAlpha())
			gABT:SetBackgroundColor(gUI:GetBackgroundColor())
			gABT:SetNormalColor(gUI:GetBackdropBorderColor())
		end

		gUI.CreateUIShadow = function(self, object, size, r, g, b, a)
			self:argCheck(object, 1, "table")
			self:argCheck(r, 2, "number", "nil")
			self:argCheck(g, 3, "number", "nil")
			self:argCheck(b, 4, "number", "nil")
			self:argCheck(a, 5, "number", "nil")
			if not(size) or (size == "small") then
				return templates["smallglow"](object, r, g, b, a)
			elseif (size == "large") then
				return templates["largeglow"](object, r, g, b, a)
			elseif (size == "huge") then
				return templates["hugeglow"](object, r, g, b, a)
			elseif (size == "tiny") then
				return templates["tinyglow"](object, r, g, b, a)
			elseif (templates[size]) then
				return templates[size](object, r, g, b, a)
			end
		end
		
		gUI.SetUIShadowColor = function(self, object, r, g, b, a)
			self:argCheck(object, 1, "table")
			self:argCheck(r, 2, "number", "nil")
			self:argCheck(g, 3, "number", "nil")
			self:argCheck(b, 4, "number", "nil")
			self:argCheck(a, 5, "number", "nil")
			if (glowFrames[object]) then
				glowFrames[object]:SetBackdropBorderColor(r or 0, g or 0, b or 0, a or 0.75)
			end
		end
		
		gUI.GetUIShadow = function(self, object)
			self:argCheck(object, 1, "table")
			return glowFrames[object]
		end

		gUI.CreateHighlight = function(self, object, top, left, bottom, right)
			local highlight = object:CreateTexture(nil, "OVERLAY")
			highlight:ClearAllPoints()
			highlight:SetPoint("TOPLEFT", object, "TOPLEFT", (left or 0), (top or 0))
			highlight:SetPoint("BOTTOMRIGHT", object, "BOTTOMRIGHT", (right or 0), (bottom or 0))
			highlight:SetTexture(C["index"][1], C["index"][2], C["index"][3], 1/3)
			object:SetHighlightTexture(highlight)
			
			local name = object:GetName()
			if (name) then
				_G[name .. "Highlight"] = highlight
			end
			
			return object:GetHighlightTexture()
		end

		gUI.CreateChecked = function(self, object, top, left, bottom, right)
			local checked = object:CreateTexture(nil, "OVERLAY")
			checked:ClearAllPoints()
			checked:SetPoint("TOPLEFT", object, "TOPLEFT", (left or 0), (top or 0))
			checked:SetPoint("BOTTOMRIGHT", object, "BOTTOMRIGHT", (right or 0), (bottom or 0))
			checked:SetTexture(C["index"][1], C["index"][2], C["index"][3], 1/2)
			object:SetCheckedTexture(checked)
			
			local name = object:GetName()
			if (name) then
				_G[name .. "Checked"] = checked
			end

			return checked
		end

		gUI.CreatePushed = function(self, object, top, left, bottom, right)
			local pushed = object:CreateTexture(nil, "OVERLAY")
			pushed:ClearAllPoints()
			pushed:SetPoint("TOPLEFT", object, "TOPLEFT", (left or 0), (top or 0))
			pushed:SetPoint("BOTTOMRIGHT", object, "BOTTOMRIGHT", (right or 0), (bottom or 0))
			pushed:SetTexture(C["value"][1], C["value"][2], C["value"][3], 1/4)
			object:SetPushedTexture(pushed)
			
			local name = object:GetName()
			if (name) then
				_G[name .. "Pushed"] = pushed
			end
			
			return object:GetPushedTexture()
		end
				
		gUI.HookAddOn = function(self, addon, func)
			if not(addon) or not(func) then return end
			if (IsAddOnLoaded(addon)) then
				func()
			else
				addonQueue[addon] = func
			end
		end
	end
end

local me = {
	Draenor = {
		Goldpaw = true;
		Aleman = true;
		Blitzee = true;
		Bottie = true;
		Caddie = true;
		Midgarsgorm = true;
		Pandalane = true;
		Pinkcuffs = true;
		Raindancer = true;
		Sunstreak = true;
		Worgnina = true;
	};
}
gUI.IsMe = function(self)
	return (me[self:GetAttribute("realm")]) and (me[self:GetAttribute("realm")][self:GetAttribute("name")])
end

gUI.GetGold = function(self)
	
end

-- 
-- all-in-one function that retrieves all databases, set up module defaults and returns current options set
--	intended for sub-modules and external addons
--
-- gUI:GetEnvironment([object, objectDefaults])
-- 	@param object <table> the 'self' object of the module that calls it
-- 	@param objectDefaults <table> the defaults table for the module that calls it
-- 	@return L, C, F, M[, db] <tables> the current locale, the color database, the function library, media library and the module db
-- 		*while L, C and F are all tables, M is a function call and should have the format local media = M(type, name)
-- 		*the module db is the current settings of the given 'self' module based on char, realm, faction. 
gUI.GetEnvironment = function(self, object, objectDefaults, noClean)
	if (object) and (objectDefaults) then
		return L, C, F, M, object:InitializeCurrentOptionsSet(self:GetAttribute("savedvariablesglobal"), objectDefaults, noClean)
	else
		return L, C, F, M
	end
end

-- gUI.PostUpdateGUI = function(self) -- don't need this... just fire self:RefreshBlizzardOptionsMenu() directly
	-- self:RefreshBlizzardOptionsMenu()
-- end

--------------------------------------------------------------------------------------------------
--		Startup and Initialization
--------------------------------------------------------------------------------------------------
gUI.PostUpdateSettings = function(self)
end

gUI.OnInit = function(self)
	-- get the settings for the root addon on this char
	db = self:InitializeCurrentOptionsSet(self:GetAttribute("savedvariablesglobal"), defaults) 
	if (db.temporaryDisabled) then
		print(L["%s has been temporarily disabled. Type /enablegui to enable it."]:format(gUI:GetAttribute("title")))
		self:Kill()
		return 
	end
	
	-- set up gActionButtons-3.0s global styling func,
	-- as this is available to all 
	do
		gUI:SetUpButtonStyling()
	end
	
	-- init modules according to settings
	do
		local module
		for name, enabled in pairs(db.enabledModules) do
			if (enabled ~= 0) then
				module = self:GetModule(name, true) 
				if (module) then
					module:Init()
				end
			end
		end
	end

	-- create the root optionsmenu for the addon
	do
		RegisterMenuWidgets() -- register our cool custom styled widgets
		
		-- we need to iterate through the modules here
		local restoreDefaults = function()
			if (InCombatLockdown()) then 
				print(L["Can not apply default settings while engaged in combat."])
				return
			end
			
			-- reset our own options
			self:ResetCurrentOptionsSetToDefaults()

			-- iterate all our children for common reset functions
			for name, child in self:IterateModules() do
				if (child.RestoreDefaults) then
					child:RestoreDefaults()
				elseif (child.ResetCurrentOptionsSetToDefaults) then
					child:ResetCurrentOptionsSetToDefaults()
				end
			end
		end
		self:RegisterAsBlizzardOptionsMenu(self:GetModule("Bags"):GetOptionsTable(), gUI:GetAttribute("title"), "default", restoreDefaults)
	end

	self:PostUpdateSettings() -- initial update of elements according to saved settings
end

gUI.OnEnable = function(self)
	db.gold = GetMoney() -- retrieve current gold on login

	-- enable modules according to settings
	do
		local module
		for name, enabled in pairs(db.enabledModules) do
			if (enabled ~= 0) then
				module = self:GetModule(name, true) 
				if (module) then
					module:Enable()
				end
			end
		end
	end
end

gUI.OnDisable = function(self)
	db.gold = GetMoney() -- store current gold on logout
end
