﻿local addon, ns = ...;

ns[addon] = gLib:newAddon( addon );

gChat = ns[addon];

gChat:RegisterEvent( "PLAYER_ENTERING_WORLD" );

gChat:Slash( "menu", function() InterfaceOptionsFrame_OpenToCategory( gChat.optionspanel.name ); end, { "gchat" } );
gChat:Slash( "telltarget", function() gChat:TellTarget( msg, "target" ); end, { "tt", "wt" } );
gChat:Slash( "tellfocus", function() gChat:TellTarget( msg, "focus" ); end, { "tf", "wf" } );

local L = gChat:GetLocale();

local _G = _G;
local VERSION, ADDON = gChat.VERSION, gChat.ADDON;

local defaults = {
	-- non-visible options
	chatlinks 					= true;			-- change Prat chatlinks into clickable links
	chatbottombutton 			= true;			-- always show the bottombutton, but only when not being at the bottom

	-- other options
	boxarrows 					= true;			-- allow the use of the arrowkeys in the inputbox
	boxposition 				= false;		-- move the inputbox to the top
	chatfade 					= true;			-- whether or not the chat will fade out by itself
	cluttervisibility 			= 2; 			-- 0 = hidden, 1 = shown, 2 = hover
	clutterframeposition 		= 0; 			-- 0 = outside, 1 = inside
	clutterscreenposition 		= 0; 			-- 0 = closest to the center of the screen, 1 = closest to the edges, 2 = always right, 3 = always left
	ebmode 						= 1;			-- 0 = hidden, 1 = blizzard, 2 = clean style
	fontmode 					= 1; 			-- 0 = default, 1 = outlined, 2 = outlined + shadow
	shortmode 					= 1; 			-- 0 = no abbreviations, 1 = channelnames, loot, xp, rep, 2 = remove brackets as well
	tabstyle 					= 1;			-- 0 = Blizzard style, 1 = clean style
}

gChat_DB = gChat_DB or {};

local CHATFRAME_EDITBOX_TEXTURES = {
	["Left"] 				= ChatFrame1EditBoxLeft:GetTexture();
	["Right"] 				= ChatFrame1EditBoxRight:GetTexture();
	["Mid"] 				= ChatFrame1EditBoxMid:GetTexture();
	["FocusLeft"] 			= ChatFrame1EditBoxFocusLeft:GetTexture();
	["FocusMid"] 			= ChatFrame1EditBoxFocusMid:GetTexture();
	["FocusRight"] 			= ChatFrame1EditBoxFocusRight:GetTexture();
};

local CHATFRAME_TAB_TEXTURES =  {
	["Left"] 				= ChatFrame1TabLeft:GetTexture();
	["Middle"] 				= ChatFrame1TabMiddle:GetTexture();
	["Right"] 				= ChatFrame1TabRight:GetTexture();
	["SelectedLeft"] 		= ChatFrame1TabSelectedLeft:GetTexture();
	["SelectedMiddle"] 		= ChatFrame1TabSelectedMiddle:GetTexture();
	["SelectedRight"] 		= ChatFrame1TabSelectedRight:GetTexture();
	["HighlightLeft"] 		= ChatFrame1TabHighlightLeft:GetTexture();
	["HighlightMiddle"] 	= ChatFrame1TabHighlightMiddle:GetTexture();
	["HighlightRight"] 		= ChatFrame1TabHighlightRight:GetTexture();
};

_FCF_SetButtonSide = function(frame, side, forceUpdate)

	if (( frame.buttonSide == side  ) and (oldmode == cmode) and not(forceUpdate) ) then
		return;
	end

	local cmode = gChat_DB.clutterframeposition
	local oldmode = frame.cluttermode
	local p = 6;
	local topY = 0;

	if ( IsCombatLog(frame) ) then
		topY = topY + CombatLogQuickButtonFrame_Custom:GetHeight();
	end
	
	frame.buttonFrame:ClearAllPoints();

	if ( side == "left" ) then
		if (cmode == 0) then
			-- outside the frame!
			frame.buttonFrame:SetPoint("TOPRIGHT", frame, "TOPLEFT", -p, topY);
			frame.buttonFrame:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", -p, 0);
		else
			-- inside the frame?
			frame.buttonFrame:SetPoint("TOPLEFT", frame, "TOPLEFT", 0, topY);
			frame.buttonFrame:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", 0, 0);
		end
	elseif ( side == "right" ) then
		if (cmode == 0) then
			-- outside the frame!
			frame.buttonFrame:SetPoint("TOPLEFT", frame, "TOPRIGHT", p, topY);
			frame.buttonFrame:SetPoint("BOTTOMLEFT", frame, "BOTTOMRIGHT", p, 0);
		else
			-- inside the frame?
			frame.buttonFrame:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 0, topY);
			frame.buttonFrame:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", 0, 0);
		end
	end

	if ( frame == DEFAULT_CHAT_FRAME ) then
		ChatFrameMenu_UpdateAnchorPoint();
	end
	
	frame.cluttermode = cmode;
	frame.buttonSide = side;
end
hooksecurefunc( "FCF_SetButtonSide", _FCF_SetButtonSide );

_FCF_UpdateButtonSide = function(chatframe)
	if not(chatframe) then return end
	
	local leftDist = chatframe:GetLeft();
	local rightDist = GetScreenWidth() - chatframe:GetRight();
	local changed = nil;
	local cmode = gChat_DB.clutterscreenposition;
	
	-- at the center
	if (cmode == 0) then
		if (leftDist <= rightDist) then 
			if (chatframe.buttonSide ~= "right") then
				FCF_SetButtonSide(chatframe, "right");
				changed = 1;
			end
		elseif (chatframe.buttonSide ~= "left") then
			FCF_SetButtonSide(chatframe, "left");
			changed = 1;
		end

		-- at the edges
	elseif (cmode == 1) then
		if (leftDist >= rightDist) then 
			if (chatframe.buttonSide ~= "right") then
				FCF_SetButtonSide(chatframe, "right");
				changed = 1;
			end
		elseif (chatframe.buttonSide ~= "left") then
			FCF_SetButtonSide(chatframe, "left");
			changed = 1;
		end

		-- always right
	elseif (cmode == 2) then
		if (chatframe.buttonSide ~= "right") then
			FCF_SetButtonSide(chatframe, "right");
			changed = 1;
		end
	
	-- always left
	elseif (cmode == 3) then
		if (chatframe.buttonSide ~= "left") then 
			FCF_SetButtonSide(chatframe, "left");
			changed = 1;
		end

	end
	return changed;
end
hooksecurefunc( "FCF_UpdateButtonSide", _FCF_UpdateButtonSide );

_FCF_FadeInChatFrame = function(chatFrame)
	local frameName = chatFrame:GetName();
	chatFrame.hasBeenFaded = true;
	for index, value in pairs(CHAT_FRAME_TEXTURES) do
		local object = _G[frameName..value];
		if ( object:IsShown() ) then
			UIFrameFadeIn(object, CHAT_FRAME_FADE_TIME, object:GetAlpha(), max(chatFrame.oldAlpha, DEFAULT_CHATFRAME_ALPHA));
		end
	end
	if ( chatFrame == FCFDock_GetSelectedWindow(GENERAL_CHAT_DOCK) ) then
		for _, frame in pairs(FCFDock_GetChatFrames(GENERAL_CHAT_DOCK)) do
			if ( frame ~= chatFrame ) then
				FCF_FadeInChatFrame(frame);
			end
		end
		if ( GENERAL_CHAT_DOCK.overflowButton:IsShown() ) then
			UIFrameFadeIn(GENERAL_CHAT_DOCK.overflowButton, CHAT_FRAME_FADE_TIME, GENERAL_CHAT_DOCK.overflowButton:GetAlpha(), CHAT_FRAME_TAB_SELECTED_MOUSEOVER_ALPHA);
		end
	end
	
	local chatTab = _G[frameName.."Tab"];
	UIFrameFadeIn(chatTab, CHAT_FRAME_FADE_TIME, chatTab:GetAlpha(), chatTab.mouseOverAlpha);
end
hooksecurefunc( "FCF_FadeInChatFrame", _FCF_FadeInChatFrame );

_FCF_FadeOutChatFrame = function(chatFrame)
	local frameName = chatFrame:GetName();
	chatFrame.hasBeenFaded = nil;
	for index, value in pairs(CHAT_FRAME_TEXTURES) do
		-- Fade out chat frame
		local object = _G[frameName..value];
		if ( object:IsShown() ) then
			UIFrameFadeOut(object, CHAT_FRAME_FADE_OUT_TIME, max(object:GetAlpha(), chatFrame.oldAlpha), chatFrame.oldAlpha);
		end
	end
	if ( chatFrame == FCFDock_GetSelectedWindow(GENERAL_CHAT_DOCK) ) then
		for _, frame in pairs(FCFDock_GetChatFrames(GENERAL_CHAT_DOCK)) do
			if ( frame ~= chatFrame ) then
				FCF_FadeOutChatFrame(frame);
			end
		end
		if ( GENERAL_CHAT_DOCK.overflowButton:IsShown() ) then
			UIFrameFadeOut(GENERAL_CHAT_DOCK.overflowButton, CHAT_FRAME_FADE_OUT_TIME, GENERAL_CHAT_DOCK.overflowButton:GetAlpha(), CHAT_FRAME_TAB_SELECTED_NOMOUSE_ALPHA);
		end
	end
	
	local chatTab = _G[frameName.."Tab"];
	UIFrameFadeOut(chatTab, CHAT_FRAME_FADE_OUT_TIME, chatTab:GetAlpha(), chatTab.noMouseAlpha);
end
hooksecurefunc( "FCF_FadeOutChatFrame", _FCF_FadeOutChatFrame );

_FloatingChatFrame_OnMouseScroll = function(self, delta)
	if (delta < 0) then
		if IsShiftKeyDown() then
			self:ScrollToBottom();
		--[[
		else
			for i = 1, 3 do
				self:ScrollDown();
			end
		]]--
		end
	elseif (delta > 0) then
		if IsShiftKeyDown() then
			self:ScrollToTop();
		--[[
		else
			for i = 1, 3 do
				self:ScrollUp();
			end
		]]--
		end
	end
end
hooksecurefunc( "FloatingChatFrame_OnMouseScroll", _FloatingChatFrame_OnMouseScroll );

local function ghettofixin( frame, fadetime, startalpha, targetalpha )
	if not frame:GetName() then return end
	if strfind( frame:GetName(), "ButtonFrame" ) then
		if ( gChat_DB.cluttervisibility == 0 ) then
			UIFrameFadeOut( frame, CHAT_FRAME_FADE_OUT_TIME, frame:GetAlpha(), CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA );
		end
	end
end
hooksecurefunc( "UIFrameFadeIn", ghettofixin );

local function ghettofixout( frame, fadetime, startalpha, targetalpha )
	if not frame:GetName() then return end
	if strfind( frame:GetName(), "ButtonFrame" ) then
		if ( gChat_DB.cluttervisibility == 1 ) then
			UIFrameFadeIn( frame, CHAT_FRAME_FADE_OUT_TIME, frame:GetAlpha(), 1 );
		end
	end
end
hooksecurefunc( "UIFrameFadeOut", ghettofixout );

local function ChatAddMessage(self, msg, ...)
	if (msg == nil) then return; end
	
	if not(gChat_DB.shortmode == 0) then
		-- this is for numbered channels only
		msg = gsub(msg, "|Hchannel:(%w+):(%d)|h%[(%d)%. (%w+)%]|h", function(a, an, b, c)
			if not(L["CHANNELS"]) then L["CHANNELS"] = {} end
			if not(L["CHANNELS"][c]) then
				L["CHANNELS"][c] = c;
			end   
			return "|Hchannel:"..a..":"..an.."|h"..b.."."..L["CHANNELS"][c]:upperonly().."|h";
		end)

		-- abbreviate the achievement spam
		-- did not find a global string for this :(
		msg = gsub(msg, L["FULL_ACHIEVEMENT_GAINED"], 		L["SHORT_ACHIEVEMENT_GAINED"]);
	end
	
	if (gChat_DB.shortmode == 2) then
		-- remove brackets
		msg = gsub(msg, 	"|Hchannel:(%w+)|h%[(%w+)%]|h", 			"|Hchannel:%1|h%2|h"		);
		msg = gsub(msg, 	"|Hchannel:(%w+)|h%[(%d)%. (%w+)%]|h", 		"|Hchannel:%1|h%2.%3|h"		);
		msg = gsub(msg, 	"|Hplayer:(.+):(.+)|h%[(.+)%]|h%:", 		"|Hplayer:%1:%2|h%3|h:"		);
	end
	
	if (gChat_DB.chatlinks) then
		-- Makes Prat links clickable without having Prat installed
		-- Original patterns retrieved from Prat 3.0 RC8 by Sylvanaar, Dec 1st 2008. (Slightly modified to fit gChat)
		msg = gsub (msg, "{CLINK:achievement:(%x+):(%-?%d-:%x-:%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-):([^}]-)}", "|c%1|Hachievement:%2|h[%3]|h|r");
		msg = gsub (msg, "{CLINK:trade:(%x+):(%-?%d-:%-?%d-:%-?%d-:%x-:[^}:]+):([^}]-)}","|c%1|Htrade:%2|h[%3]|h|r");
		msg = gsub (msg, "{CLINK:quest:(%x+):(%-?%d-):(%-?%d-):([^}]-)}", "|c%1|Hquest:%2:%3|h%4|h|r");
		msg = gsub (msg, "{CLINK:item:(%x+):(%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-:%-?%d-):([^}]-)}", "|c%1|Hitem:%2|h%[%3%]|h|r");
		msg = gsub (msg, "{CLINK:enchant:(%x+):(%-?%d-):([^}]-)}", "|c%1|Henchant:%2|h%[%3%]|h|r");
		msg = gsub (msg, "{CLINK:spell:(%x+):(%-?%d-):([^}]-)}", "|c%1|Hspell:%2|h%[%3%]|h|r");
	end
	
	self:BlizzAddMessage(msg, ...);
end

local function SetAbbreviations()
	-- if our local copies of the original defaults aren't yet made, make them!
	if not(L["DEFAULT_ABBREVIATIONS"]) then
		L["DEFAULT_ABBREVIATIONS"] = {}
		for i,v in pairs(L["ABBREVIATIONS"]) do
			L["DEFAULT_ABBREVIATIONS"][i] = _G[i];
		end
	end
	
	if (gChat_DB.shortmode == 0) then
		-- set the globals back to their defaults
		for i,v in pairs(L["ABBREVIATIONS"]) do
			_G[i] = L["DEFAULT_ABBREVIATIONS"][i];
		end
	else
		-- replace the dreaded globals with our versions
		for i,v in pairs(L["ABBREVIATIONS"]) do
			_G[i] = L["ABBREVIATIONS"][i];
		end
	end
end
local function SetArrowKey()
	for _,name in pairs(CHAT_FRAMES) do 
		_G[name.."EditBox"]:SetAltArrowKeyMode(not gChat_DB.boxarrows);
	end
end
local function SetFade()
	for _,name in pairs(CHAT_FRAMES) do 
		_G[name]:SetFading(gChat_DB.chatfade);
	end
end
local function SetSticky()
	-- make pretty much everything sticky, so you can press ENTER to directly speak in Trade, Officer, etc etc
	-- this is marked for configuration and saved settings in a future version of gChat
	local stickies = {"SAY", "PARTY", "RAID", "RAID_WARNING", "GUILD", "OFFICER", "BATTLEGROUND", "CHANNEL", "CHANNEL1", "CHANNEL2", "CHANNEL3", "CHANNEL4", "CHANNEL5", "CHANNEL6", "CHANNEL7", "CHANNEL8", "CHANNEL9", "CHANNEL10"}
	for i,v in pairs(stickies) do 
		ChatTypeInfo[v].sticky = 1; 
	end
end

local function SetEditBoxPosition()
	for _,name in pairs(CHAT_FRAMES) do 
		if (gChat_DB.boxposition) then
			-- editBox at the top
			_G[name.."EditBox"]:ClearAllPoints();
			_G[name.."EditBox"]:SetPoint("BOTTOMLEFT", _G[name], "TOPLEFT", -5, 2);
			_G[name.."EditBox"]:SetPoint("BOTTOMRIGHT", _G[name], "TOPRIGHT", 5, 2);
		else
			-- default position, at the bottom
			_G[name.."EditBox"]:ClearAllPoints();
			_G[name.."EditBox"]:SetPoint("TOPLEFT", _G[name], "BOTTOMLEFT", -5, -2);
			_G[name.."EditBox"]:SetPoint("TOPRIGHT", _G[name], "BOTTOMRIGHT", 5, -2);
		end
	end
end
local function SetEditBoxStyle()
	local function SetAlphaState(f)
		if (gChat_DB.ebmode == 1) then
			-- default box, show it
			for name, texture in pairs(CHATFRAME_EDITBOX_TEXTURES) do
				_G[f:GetName().."EditBox"..name]:SetTexture(texture);
			end
		else
			-- hide the default box
			for name, texture in pairs(CHATFRAME_EDITBOX_TEXTURES) do
				_G[f:GetName().."EditBox"..name]:SetTexture("");
			end
		end
	end
	for _,name in pairs(CHAT_FRAMES) do 
		SetAlphaState(_G[name]);
	end
end
local function SetFontStyle()
	local tabfontheight, editboxfontheight = 12, 14;

	for _,name in pairs(CHAT_FRAMES) do 
		local fontset, fontheight, fontflags = _G[name]:GetFont();

		_G[name]:SetShadowColor( 0, 0, 0, 0.75 );  
		_G[name.."TabText"]:SetShadowColor( 0, 0, 0, 0.75 );
		_G[name.."EditBox"]:SetShadowColor( 0, 0, 0, 0.75 );
		_G[name.."EditBoxHeader"]:SetShadowColor( 0, 0, 0, 0.75 );

		-- blizzard style
		if (gChat_DB.fontmode == 0) then
			_G[name]:SetFont( fontset, fontheight , "" );
			_G[name.."TabText"]:SetFont( fontset, tabfontheight, "" );
			_G[name.."EditBox"]:SetFont( fontset, editboxfontheight, "" );
			_G[name.."EditBoxHeader"]:SetFont( fontset, editboxfontheight, "" );

			_G[name]:SetShadowOffset( 1.25, -1.25 ); 
			_G[name.."TabText"]:SetShadowOffset( 1.25, -1.25 );
			_G[name.."EditBox"]:SetShadowOffset( 1.25, -1.25 ); 
			_G[name.."EditBoxHeader"]:SetShadowOffset( 1.25, -1.25 ); 

		-- clean style
		elseif (gChat_DB.fontmode == 1) then
			_G[name]:SetFont( fontset, fontheight , "OUTLINE" );
			_G[name.."TabText"]:SetFont( fontset, tabfontheight, "OUTLINE" );
			_G[name.."EditBox"]:SetFont( fontset, editboxfontheight, "OUTLINE" );
			_G[name.."EditBoxHeader"]:SetFont( fontset, editboxfontheight, "OUTLINE" );

			_G[name]:SetShadowOffset( 1.25, -1.25 ); 
			_G[name.."TabText"]:SetShadowOffset( 1.25, -1.25 );
			_G[name.."EditBox"]:SetShadowOffset( 1.25, -1.25 ); 
			_G[name.."EditBoxHeader"]:SetShadowOffset( 1.25, -1.25 ); 

		-- cartoon style
		elseif (gChat_DB.fontmode == 2) then
			_G[name]:SetFont( fontset, fontheight , "OUTLINE" );
			_G[name.."TabText"]:SetFont( fontset, tabfontheight, "OUTLINE" );
			_G[name.."EditBox"]:SetFont( fontset, editboxfontheight, "OUTLINE" );
			_G[name.."EditBoxHeader"]:SetFont( fontset, editboxfontheight, "OUTLINE" );

			_G[name]:SetShadowOffset( 2, -2 ); 
			_G[name.."TabText"]:SetShadowOffset( 2, -2 );
			_G[name.."EditBox"]:SetShadowOffset( 2, -2 ); 
			_G[name.."EditBoxHeader"]:SetShadowOffset( 2, -2 ); 

		end
	end
end
local function SetTabStyle()
	for _,frame in pairs(CHAT_FRAMES) do

		-- default Blizzard style
		if (gChat_DB.tabstyle == 0) then
			for name, texture in pairs(CHATFRAME_TAB_TEXTURES) do
				_G[frame.."Tab"..name]:SetTexture(texture);

				CHAT_FRAME_FADE_TIME = 2;
				CHAT_FRAME_FADE_OUT_TIME = 2;
				CHAT_TAB_SHOW_DELAY = 1;
				CHAT_TAB_HIDE_DELAY = 1;
				CHAT_FRAME_TAB_SELECTED_MOUSEOVER_ALPHA = 1;
				CHAT_FRAME_TAB_SELECTED_NOMOUSE_ALPHA = 0.4;
				CHAT_FRAME_TAB_NORMAL_MOUSEOVER_ALPHA = 0.6;
				CHAT_FRAME_TAB_NORMAL_NOMOUSE_ALPHA = 0.2;
				CHAT_FRAME_TAB_ALERTING_MOUSEOVER_ALPHA = 1;
				CHAT_FRAME_TAB_ALERTING_NOMOUSE_ALPHA = 1;
				CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA = 0.2;
				DEFAULT_CHATFRAME_ALPHA = 0.25;	
				
			end

		-- gUI style
		elseif (gChat_DB.tabstyle == 1) then
			for name, texture in pairs(CHATFRAME_TAB_TEXTURES) do
				_G[frame.."Tab"..name]:SetTexture("");

				CHAT_FRAME_FADE_TIME = 0.5;
				CHAT_FRAME_FADE_OUT_TIME = 0.5;
				CHAT_TAB_SHOW_DELAY = 0.0;
				CHAT_TAB_HIDE_DELAY = 0.0;
				CHAT_FRAME_TAB_SELECTED_MOUSEOVER_ALPHA = 1;
				CHAT_FRAME_TAB_SELECTED_NOMOUSE_ALPHA = 0;
				CHAT_FRAME_TAB_NORMAL_MOUSEOVER_ALPHA = 0.4;
				CHAT_FRAME_TAB_NORMAL_NOMOUSE_ALPHA = 0;
				CHAT_FRAME_TAB_ALERTING_MOUSEOVER_ALPHA = 1;
				CHAT_FRAME_TAB_ALERTING_NOMOUSE_ALPHA = 0;
				CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA = 0.0;
				DEFAULT_CHATFRAME_ALPHA = 0.0;
				
			end
		end
	end
end
local function SetClutterDisplayMode()
	local func = {
		[0] = function(frame) UIFrameFadeOut(frame, CHAT_FRAME_FADE_OUT_TIME, frame:GetAlpha(), CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA); end;
		[1] = function(frame) UIFrameFadeIn(frame, CHAT_FRAME_FADE_TIME, frame:GetAlpha(), 1); end;
		[2] = function(frame) UIFrameFadeOut(frame, CHAT_FRAME_FADE_OUT_TIME, frame:GetAlpha(), CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA); end;
	}

	for _,name in pairs(CHAT_FRAMES) do 
		func[gChat_DB.cluttervisibility]( _G[name.."ButtonFrame"] );
	end

	func[gChat_DB.cluttervisibility]( ChatFrameMenuButton );
	func[gChat_DB.cluttervisibility]( FriendsMicroButton );
end

local function UpdateAllButtonSides()
	for _,name in pairs(CHAT_FRAMES) do 
		f = _G[name];
		FCF_UpdateButtonSide(f);
		FCF_SetButtonSide(f, FCF_GetButtonSide(f));
	end
end
local function UpdateAll()
	
	SetAbbreviations();
	SetArrowKey();
	SetFade();
	SetSticky();
	
	SetEditBoxPosition();
	SetEditBoxStyle();
	SetFontStyle();
	SetTabStyle();

	SetClutterDisplayMode();
	UpdateAllButtonSides();

end

local function OnChatHover( frame, elapsed )
	frame.elapsed = ( frame.elapsed or 0 ) + elapsed

	if ( frame.elapsed < 0.1 ) then
		return
	end

	frame.elapsed = frame.elapsed - 0.1;

	if ( gChat_DB.cluttervisibility == 2 ) then
		if ( MouseIsOver( frame ) or MouseIsOver( frame.buttonFrame ) ) then
			UIFrameFadeIn( frame.buttonFrame, CHAT_FRAME_FADE_TIME, frame.buttonFrame:GetAlpha(), 1);
			if ( frame == DEFAULT_CHAT_FRAME ) then 
				UIFrameFadeIn( ChatFrameMenuButton, CHAT_FRAME_FADE_TIME, ChatFrameMenuButton:GetAlpha(), 1 );
				UIFrameFadeIn( FriendsMicroButton, CHAT_FRAME_FADE_TIME, FriendsMicroButton:GetAlpha(), 1 );
			elseif ( frame.isDocked ) then
				UIFrameFadeIn( FriendsMicroButton, CHAT_FRAME_FADE_TIME, FriendsMicroButton:GetAlpha(), 1 );
			end
		else
			UIFrameFadeOut( frame.buttonFrame, CHAT_FRAME_FADE_OUT_TIME, frame.buttonFrame:GetAlpha(), CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA);
			if ( frame == DEFAULT_CHAT_FRAME ) then 
				UIFrameFadeOut( ChatFrameMenuButton, CHAT_FRAME_FADE_OUT_TIME, ChatFrameMenuButton:GetAlpha(), CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA );
				UIFrameFadeOut( FriendsMicroButton, CHAT_FRAME_FADE_OUT_TIME, FriendsMicroButton:GetAlpha(), CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA );
			elseif ( frame.isDocked ) then
				UIFrameFadeOut( FriendsMicroButton, CHAT_FRAME_FADE_OUT_TIME, FriendsMicroButton:GetAlpha(), CHAT_FRAME_BUTTON_FRAME_MIN_ALPHA );
			end
		end
	end
end

function gChat:TellTarget(txt, who)
  if UnitExists(who) and UnitIsPlayer(who) and UnitIsFriend("player", who) then
    local name = gChat:GetCharName(who)
	SendChatMessage(txt, "WHISPER", nil, name)    
    ChatEdit_SetLastToldTarget(name)	
  end
end
function gChat:TellTargetKeyBind(who)
	local editBox = DEFAULT_CHAT_FRAME.editBox
	local text = editBox:GetText()

	editBox:Show()
	editBox:SetAttribute("chatType", "WHISPER")
	if UnitExists(who) and UnitIsPlayer(who) and UnitIsFriend("player", who) then
		local name = gChat:GetCharName(who)	
		editBox:SetAttribute("tellTarget", name)
		ChatEdit_UpdateHeader(editBox)
	else
		editBox:SetAttribute("tellTarget", who)
		if( who == "target") then
			editBox:SetText("/tt ")	
		elseif( who == "focus") then
			editBox:SetText("/tf ")
		end
		ChatEdit_UpdateHeader(editBox)
	end
end
function gChat:GetCharName(who)
	local targetName, targetServerName = UnitName(who)
	
	if not targetServerName or (#targetServerName < 1) then
		return targetName
		
	elseif targetServerName ~= gChat.REALM then
	    local crossServerName = targetName.."-"..targetServerName
		return crossServerName
	end
end

function gChat:OnInit()
	gChat_DB = GL_Table_Validate( gChat_DB, defaults );

	BINDING_HEADER_GCHAT = self.NAME;
	BINDING_NAME_GCHAT_TELLTARGET = L["KEYBIND_WHISPER_TARGET"];
	BINDING_NAME_GCHAT_TELLFOCUS = L["KEYBIND_WHISPER_FOCUS"];
end
function gChat:OnEnable()
	local function SetUpChat(frame)
		local editbox = _G[frame:GetName().."EditBox"];
		local tab = _G[frame:GetName().."Tab"];

		frame:SetFrameStrata("LOW");
	
		frame:SetClampRectInsets( 0, 0, 0, 0 );
		frame:SetHitRectInsets( 0, 0, 0, 0 );
		frame:SetClampedToScreen( true );

		editbox:HookScript("OnEditFocusGained", function(self) self:Show() end)
		editbox:HookScript("OnEditFocusLost", function(self) self:Hide() end)
		editbox:Hide();

		tab:HookScript("OnClick", function() editbox:Hide() end)
		_G[frame:GetName().."ClickAnywhereButton"]:HookScript("OnClick", function() editbox:Hide() end)

		for i,v in pairs(CHAT_FRAME_TEXTURES) do
			if strfind( v, "ButtonFrame" ) then
				_G[frame:GetName()..v]:SetTexture("");
			end
		end
		
		frame:HookScript( "OnUpdate", OnChatHover );

		frame.BlizzAddMessage = frame.AddMessage;
		frame.AddMessage = ChatAddMessage;

		if not(frame.isDocked == 1) then
			tab:SetAlpha(0);
		end
		
		hooksecurefunc("FCFTab_UpdateAlpha", function()
			tab:SetAlpha(0);
			tab.mouseOverAlpha = ((frame == FCFDock_GetSelectedWindow(GENERAL_CHAT_DOCK)) or not(frame.isDocked == 1)) and CHAT_FRAME_TAB_SELECTED_MOUSEOVER_ALPHA or CHAT_FRAME_TAB_NORMAL_MOUSEOVER_ALPHA;
			tab.noMouseAlpha = CHAT_FRAME_TAB_NORMAL_NOMOUSE_ALPHA;
		end)
	end
	
	for _,name in pairs(CHAT_FRAMES) do 
		SetUpChat(_G[name]);
	end

	BNToastFrame:HookScript("OnShow", function(self)
		self:Place("BOTTOMLEFT", ChatFrame1, "TOPLEFT", 0, 16);
	end)
	
	hooksecurefunc("ChatEdit_OnSpacePressed", function()
		local editBox = DEFAULT_CHAT_FRAME.editBox;
		local text = editBox:GetText();
		local who = "target";
		
		if (string.len(text) ~= 4) then return end
		if (strsub(text, 1, 1) ~= "/") then return end
		
		local slashcmd = strsub(text, 2, 3)
		if slashcmd == "wf" or slashcmd == "tf" then	
			who = "focus"; 
		elseif slashcmd ~= "wt" and slashcmd ~= "tt" then
			return
		end
		
		editBox:SetAttribute("chatType", "WHISPER")
		if UnitExists(who) and UnitIsPlayer(who) and UnitIsFriend("player", who) then
			local name = gChat:GetCharName(who)	
			editBox:SetText("")
			editBox:SetAttribute("tellTarget", name)
			ChatEdit_UpdateHeader(editBox)
		else
			editBox:SetAttribute("tellTarget", who)
			ChatEdit_UpdateHeader(editBox)
		end
	end)

	hooksecurefunc("FCF_OpenTemporaryWindow", function()
		SetUpChat(FCF_GetCurrentChatFrame());
		UpdateAll();
	end)


 	self.optionspanel = GL_CreateInterfaceOptionsPanel({
		title = L["TITLE"];
		version = VERSION;
		description = "";
		panel = nil;
		option = {
			[40] = {
				name = L["GUI_CLUTTER_SCRPOS"];
				type = "dropdown";
				value = gChat_DB.clutterscreenposition;
				list = {
					{
						level 		= 1;
						text 		= function() 
							return ((gChat_DB.clutterscreenposition == 0) and L["GUI_CLUTTER_SCRPOS_OPTION_CENTER"]) or ((gChat_DB.clutterscreenposition == 1) and L["GUI_CLUTTER_SCRPOS_OPTION_EDGE"]) or ((gChat_DB.clutterscreenposition == 2) and L["GUI_CLUTTER_SCRPOS_OPTION_RIGHT"]) or ((gChat_DB.clutterscreenposition == 3) and L["GUI_CLUTTER_SCRPOS_OPTION_LEFT"]) or "";
						end;
						menulist = {
							{
								text 			= L["GUI_CLUTTER_SCRPOS_CENTER"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.clutterscreenposition = 0;
									UpdateAllButtonSides();
								end;
							};
							{
								text 			= L["GUI_CLUTTER_SCRPOS_EDGE"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.clutterscreenposition = 1;
									UpdateAllButtonSides();
								end;
							};
							{
								text 			= L["GUI_CLUTTER_SCRPOS_RIGHT"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.clutterscreenposition = 2;
									UpdateAllButtonSides();
								end;
							};
							{
								text 			= L["GUI_CLUTTER_SCRPOS_LEFT"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.clutterscreenposition = 3;
									UpdateAllButtonSides();
								end;
							};
						};
					};
				};
			};
			[45] = {
				name = L["GUI_CLUTTER_POSITION"];
				type = "dropdown";
				value = gChat_DB.clutterframeposition;
				list = {
					{
						level 		= 1;
						text 		= function() 
							return ((gChat_DB.clutterframeposition == 0) and L["GUI_CLUTTER_POSITION_OPTION_OUTSIDE"]) or ((gChat_DB.clutterframeposition == 1) and L["GUI_CLUTTER_POSITION_OPTION_INSIDE"]) or "";
						end;
						menulist = {
							{
								text 			= L["GUI_CLUTTER_POSITION_OUTSIDE"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.clutterframeposition = 0;
									UpdateAllButtonSides();
								end;
							};
							{
								text 			= L["GUI_CLUTTER_POSITION_INSIDE"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.clutterframeposition = 1;
									UpdateAllButtonSides();
								end;
							};
						};
					};
				};
			};
			[55] = {
				name = L["GUI_TAB_STYLE"];
				type = "dropdown";
				value = gChat_DB.tabstyle;
				list = {
					{
						level 		= 1;
						text 		= function() 
							return ((gChat_DB.tabstyle == 0) and L["GUI_TAB_STYLE_BLIZZARD"]) or ((gChat_DB.tabstyle == 1) and L["GUI_TAB_STYLE_CLEAN"]) or "";
						end;
						menulist = {
							{
								text 			= L["GUI_TAB_STYLE_OPTION_BLIZZARD"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.tabstyle = 0;
									SetTabStyle();
								end;
							};
							{
								text 			= L["GUI_TAB_STYLE_OPTION_CLEAN"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.tabstyle = 1;
									SetTabStyle();
								end;
							};
						};
					};
				};
			};
			[35] = {
				name = L["GUI_CLUTTER_VISIBILITY"];
				type = "dropdown";
				value = gChat_DB.cluttervisibility;
				list = {
					{
						level 		= 1;
						text 		= function() 
							return ((gChat_DB.cluttervisibility == 0) and L["GUI_CLUTTER_VISIBILITY_OPTION_HIDDEN"]) or ((gChat_DB.cluttervisibility == 1) and L["GUI_CLUTTER_VISIBILITY_OPTION_VISIBLE"]) or ((gChat_DB.cluttervisibility == 2) and L["GUI_CLUTTER_VISIBILITY_OPTION_HOVER"]) or "";
						end;
						menulist = {
							{
								text 			= L["GUI_CLUTTER_VISIBILITY_HIDDEN"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.cluttervisibility = 0;
									SetClutterDisplayMode();
								end;
							};
							{
								text 			= L["GUI_CLUTTER_VISIBILITY_VISIBLE"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.cluttervisibility = 1;
									SetClutterDisplayMode();
								end;
							};
							{
								text 			= L["GUI_CLUTTER_VISIBILITY_HOVER"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.cluttervisibility = 2;
									SetClutterDisplayMode();
								end;
							};
						};
					};
				};
			};
			[30] = {
				name = L["GUI_EDITBOX_MODE"];
				type = "dropdown";
				value = gChat_DB.ebmode;
				list = {
					{
						level = 1;
						text = function() 
							return ((gChat_DB.ebmode == 0) and L["GUI_EDITBOX_MODE_OPTION_HIDDEN"]) or ((gChat_DB.ebmode == 1) and L["GUI_EDITBOX_MODE_OPTION_BLIZZARD"]) or "";
						end;
						menulist = {
							{
								text = L["GUI_EDITBOX_MODE_HIDDEN"];
								notCheckable = 1;
								func = function() 
									gChat_DB.ebmode = 0;
									SetEditBoxStyle();
								end;
							};
							{
								text = L["GUI_EDITBOX_MODE_BLIZZARD"];
								notCheckable = 1;
								func = function() 
									gChat_DB.ebmode = 1;
									SetEditBoxStyle();
								end;
							};
						};
					};
				};
			};
			[10] = {
				name = L["GUI_SHORTS"];
				type = "dropdown";
				value = gChat_DB.shortmode;
				list = {
					{
						level 		= 1;
						text 		= function() 
							return ((gChat_DB.shortmode == 0) and L["GUI_SHORTS_OPTION_NONE"]) or ((gChat_DB.shortmode == 1) and L["GUI_SHORTS_OPTION_CHAN"]) or ((gChat_DB.shortmode == 2) and L["GUI_SHORTS_OPTION_CHANPLAY"]) or "";
						end;
						menulist = {
							{
								text 			= L["GUI_SHORTS_NONE"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.shortmode = 0;
									SetAbbreviations(); 
								end;
							};
							{
								text 			= L["GUI_SHORTS_CHAN"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.shortmode = 1;
									SetAbbreviations(); 
								end;
							};
							{
								text 			= L["GUI_SHORTS_CHANPLAY"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.shortmode = 2;
									SetAbbreviations(); 
								end;
							};
						};
					};
				};
			};
			[60] = {
				name = L["GUI_FONT"];
				type = "dropdown";
				value = gChat_DB.fontmode;
				list = {
					{
						level 		= 1;
						text 		= function() 
							return ((gChat_DB.fontmode == 0) and L["GUI_FONT_OPTION_NONE"]) or ((gChat_DB.fontmode == 1) and L["GUI_FONT_OPTION_OUTLINE"]) or ((gChat_DB.fontmode == 2) and L["GUI_FONT_OPTION_SHADOW"]) or "";
						end;
						menulist = {
							{
								text 			= L["GUI_FONT_NONE"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.fontmode = 0;
									SetFontStyle();
								end;
							};
							{
								text 			= L["GUI_FONT_OUTLINE"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.fontmode = 1;
									SetFontStyle();
								end;
							};
							{
								text 			= L["GUI_FONT_SHADOW"];
								notCheckable 	= 1;
								func = function() 
									gChat_DB.fontmode = 2;
									SetFontStyle();
								end;
							};
						};
					};
				};
			};
			[100] = {
				name = L["GUI_CHAT_MOVEEDITBOX"];
				type = "button";
				value = gChat_DB.boxposition;
				panelfunction = function() gChat_DB.boxposition = not gChat_DB.boxposition; SetEditBoxPosition(); end;
			};
			[120] = {
				name = L["GUI_CHAT_ENABLEARROWS"];
				type = "button";
				value = gChat_DB.boxarrows;
				panelfunction = function() gChat_DB.boxarrows = not gChat_DB.boxarrows; SetArrowKey(); end;
			};
			[130] = {
				name = L["GUI_CHAT_FADE"];
				type = "button";
				value = gChat_DB.chatfade;
				panelfunction = function() gChat_DB.chatfade = not gChat_DB.chatfade; SetFade(); end;
			};
		};
	}, gChat_DB, defaults);

end
function gChat:PLAYER_ENTERING_WORLD()
	UpdateAll();
end

